/*
 * Copyright (c) 2005-2008 Obeo
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Obeo - initial API and implementation
 *
 * Taken from org.eclipse.search.internal.ui.text.FileTreeContentProvider
 */

package fr.obeo.acceleo.gen.ui.popupMenus;

import java.util.Comparator;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IFolder;
import org.eclipse.core.resources.IResource;
import org.eclipse.jface.viewers.ITreeContentProvider;
import org.eclipse.jface.viewers.TreeViewer;
import org.eclipse.jface.viewers.Viewer;

/**
 * This class provide the tree structure to be shown in the
 * ReferencesSearchViewPage
 * 
 * @author Yvan LUSSAUD <a
 *         href="mailto:yvan.lussaud@obeo.fr">yvan.lussaud@obeo.fr</a>
 * 
 */
public class ReferencesTreeContentProvider implements ITreeContentProvider {
	/**
	 * The map connecting an element with there children
	 */
	private Map map = null;

	/**
	 * The result provided by the ReferencesSearchQuery
	 */
	private ReferencesSearchResult result;

	/**
	 * An empty array for elements with no child
	 */
	private final Object[] EMPTY_ARR = new Object[0];

	/**
	 * The TreeViewer from the ReferencesSearchViewPage
	 */
	private TreeViewer viewer = null;

	/**
	 * Constructor
	 * 
	 * @param viewer
	 *            the tree view to use to display the result
	 */
	public ReferencesTreeContentProvider(TreeViewer viewer) {
		this.viewer = viewer;
	}

	/**
	 * Refresh the tree viewer
	 */
	public void clear() {
		viewer.refresh();
	}

	/* (non-Javadoc) */
	public Object[] getChildren(Object parentElement) {
		Set children = (Set) map.get(parentElement);
		if (children == null)
			return EMPTY_ARR;
		return children.toArray();
	}

	/* (non-Javadoc) */
	public Object getParent(Object element) {
		if (element instanceof ReferenceEntry) {
			ReferenceEntry entry = (ReferenceEntry) element;
			return entry.getTemplateFile();
		} else if (element instanceof IFile) {
			return ((IFile) element).getParent();
		} else if (element instanceof IFolder) {
			return ((IFolder) element).getProject();
		}
		return null;
	}

	/* (non-Javadoc) */
	public boolean hasChildren(Object element) {
		return getChildren(element).length > 0;
	}

	/* (non-Javadoc) */
	public Object[] getElements(Object inputElement) {
		return getChildren(inputElement);
	}

	/* (non-Javadoc) */
	public void dispose() {
	}

	/* (non-Javadoc) */
	public void inputChanged(Viewer viewer, Object oldInput, Object newInput) {
		if (newInput instanceof ReferencesSearchResult) {
			initialize((ReferencesSearchResult) newInput);
		}
	}

	/**
	 * This method initialize the tree view according to the given input
	 * 
	 * @param newInput
	 *            the new input on inputChanged
	 */
	protected synchronized void initialize(ReferencesSearchResult newInput) {
		result = newInput;
		map = new HashMap();
		if (newInput != null) {
			Object entries[] = newInput.getElements();
			for (int i = 0; i < entries.length; i++) {
				insert(entries[i], false);
			}
		}
	}

	/**
	 * Insert a new element into the tree view
	 * 
	 * @param child
	 *            the element to insert
	 * @param refreshViewer
	 *            the tree viewer should be refreshed
	 */
	protected void insert(Object child, boolean refreshViewer) {
		Object parent = getParent(child);
		while (parent != null) {
			if (insertChild(parent, child)) {
				if (refreshViewer)
					viewer.add(parent, child);
			} else {
				if (refreshViewer)
					viewer.refresh(parent);
				return;
			}
			child = parent;
			parent = getParent(child);
		}
		if (insertChild(result, child)) {
			if (refreshViewer)
				viewer.add(result, child);
		}
	}

	/**
	 * Insert the given child to the given parent
	 * 
	 * @param parent
	 *            the parent
	 * @param child
	 *            the child
	 * @return true if the insertion was done
	 */
	private boolean insertChild(Object parent, Object child) {
		Set children = (Set) map.get(parent);
		if (children == null) {
			children = new TreeSet(new Comparator() {
				public int compare(Object arg0, Object arg1) {
					if (arg0 instanceof ReferenceEntry && arg1 instanceof ReferenceEntry) {
						int begin0 = ((ReferenceEntry) arg0).getPosition().b();
						int begin1 = ((ReferenceEntry) arg1).getPosition().b();

						if (begin0 > begin1) {
							return 1;
						} else if (begin0 < begin1) {
							return -1;
						} else {
							return 0;
						}
					} else if (arg0 instanceof IResource && arg1 instanceof IResource) {
						IResource res0 = (IResource) arg0;
						IResource res1 = (IResource) arg1;
						String name0 = res0.getProject().toString() + res0.getProjectRelativePath().toString() + res0.getName();
						String name1 = res1.getProject().toString() + res1.getProjectRelativePath().toString() + res1.getName();
						return name0.compareTo(name1);
					} else {
						return 0;
					}
				}
			});
			map.put(parent, children);
		}
		return children.add(child);
	}

	/**
	 * Refresh the tree viewer according to changes
	 * 
	 * @param updatedElements
	 *            array of updated elements
	 */
	public synchronized void elementsChanged(Object[] updatedElements) {
		for (int i = 0; i < updatedElements.length; i++) {
			if (result.getMatchCount(updatedElements[i]) > 0)
				insert(updatedElements[i], true);
			else
				remove(updatedElements[i], true);
		}
	}

	/**
	 * Remove an element
	 * 
	 * @param element
	 *            the element to remove
	 * @param refreshViewer
	 *            the tree viewer should be refreshed
	 */
	protected void remove(Object element, boolean refreshViewer) {
		// precondition here: fResult.getMatchCount(child) <= 0

		if (hasChildren(element)) {
			if (refreshViewer)
				viewer.refresh(element);
		} else {
			if (result.getMatchCount(element) == 0) {
				map.remove(element);
				Object parent = getParent(element);
				if (parent != null) {
					removeFromSiblings(element, parent);
					remove(parent, refreshViewer);
				} else {
					removeFromSiblings(element, result);
					if (refreshViewer)
						viewer.refresh();
				}
			} else {
				if (refreshViewer) {
					viewer.refresh(element);
				}
			}
		}
	}

	/**
	 * Remove the given element from the chidlren of parent
	 * 
	 * @param element
	 *            the element to remove
	 * @param parent
	 *            the parent node to use
	 */
	private void removeFromSiblings(Object element, Object parent) {
		Set siblings = (Set) map.get(parent);
		if (siblings != null) {
			siblings.remove(element);
		}
	}

}
