/*
 * Copyright (c) 2005-2008 Obeo
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Obeo - initial API and implementation
 */
package org.acceleo.gen.ui.wizards.export.model;

import java.io.File;
import java.lang.reflect.Type;

import org.eclipse.core.resources.IFile;
import org.eclipse.emf.ecore.EPackage;

/**
 * Export utility class which contains information about an EMT file.
 * 
 * @author www.obeo.fr
 * 
 */
public class AcceleoExportEMTModel {

	/** EMT file path (".emt") */
	private String emtFile;

	/** URI of the metamodel */
	private String nsURI;

	/** Flag which indicates if we use an umlModel */
	private boolean umlModel;

	/** Flag which indicates if we use as ecoreModel */
	private boolean ecoreModel;

	/**
	 * The template file (".mt"). An EMT file is the result of the compilation
	 * of this template.
	 */
	private IFile mtSrc;

	/**
	 * Gets the EMT file path (".emt").
	 * 
	 * @return the EMT File path
	 */
	public String getEmtFile() {
		return emtFile;
	}

	/**
	 * Sets the EMT file path.
	 * 
	 * @param emtFile
	 *            is the EMT file path
	 */
	public void setEmtFile(String emtFile) {
		this.emtFile = emtFile;
	}

	/**
	 * Gets the URI of the metamodel.
	 * 
	 * @return the URI of the metamodel
	 */
	public String getNsURI() {
		return nsURI;
	}

	/**
	 * Sets the URI of the metamodel
	 * 
	 * @param nsURI
	 *            is the new URI of the metamodel
	 */
	public void setNsURI(String nsURI) {
		this.nsURI = nsURI;
	}

	/**
	 * Gets the template file (".mt").
	 * 
	 * @return the template file
	 */
	public IFile getMtSrc() {
		return mtSrc;
	}

	/**
	 * Sets the template file.
	 * 
	 * @param mtSrc
	 *            is the template file
	 */
	public void setMtSrc(IFile mtSrc) {
		this.mtSrc = mtSrc;
	}

	/**
	 * Indicates if we are working in a UML context.
	 * 
	 * @return true if we are working in a UML context
	 */
	public boolean isUmlModel() {
		return umlModel;
	}

	/**
	 * Are We working in a UML context?
	 * 
	 * @param umlModel
	 *            is true if we are working in a UML context
	 */
	public void setUmlModel(boolean umlModel) {
		this.umlModel = umlModel;
	}

	/**
	 * Indicates if we are working in an ECORE context.
	 * 
	 * @return true if we are working in an ECORE context
	 */
	public boolean isEcoreModel() {
		return ecoreModel;
	}

	/**
	 * Are We working in an ECORE context?
	 * 
	 * @param ecoreModel
	 *            is true if we are working in an ECORE context
	 */
	public void setEcoreModel(boolean ecoreModel) {
		this.ecoreModel = ecoreModel;
	}

	/**
	 * Computes the EMT name. For example, it returns "Class" if the EMT file is
	 * "/org/acceleo/class.emt"
	 * 
	 * @return the EMT name
	 */
	public String getEmtName() {
		String ret = "";
		if (emtFile != null && emtFile.length() > 0) {
			String emtFileName = emtFile.substring(emtFile.lastIndexOf("/") + 1, emtFile.length());
			ret = emtFileName.substring(0, emtFileName.indexOf('.'));
			ret = ret.substring(0, 1).toUpperCase() + ret.substring(1, ret.length());
			// We remove all _ char and we capitalize next character
			for (int index = 0; index < ret.length(); index++) {
				if (!Character.isJavaIdentifierPart(ret.charAt(index))) {
					if (index == ret.length() - 1) {
						ret = ret.substring(0, index);
					} else {
						String suffix = ret.substring(index + 1);
						suffix = suffix.substring(0, 1).toUpperCase() + suffix.substring(1);
						ret = ret.substring(0, index) + suffix;
					}
				}
			}
		}
		return ret;
	}

	/**
	 * Returns the model package based upon the metamodel URI.
	 * 
	 * @return the model package based upon the metamodel URI
	 */
	public String getModelPackageName() {
		String res = null;
		if (nsURI.endsWith(".ecore")) {
			res = nsURI.substring(nsURI.lastIndexOf("/") + 1, nsURI.indexOf(".ecore"));
		} else {
			EPackage ePkg = (EPackage) EPackage.Registry.INSTANCE.get(nsURI);
			Type[] packageInterfaces = ePkg.getClass().getGenericInterfaces();
			if (packageInterfaces != null && packageInterfaces.length > 0) {
				res = ((Class) packageInterfaces[0]).getCanonicalName();
			}
		}
		return res;
	}

	/**
	 * Returns the package name.
	 * 
	 * @return the package name
	 */
	public String getPackageName() {
		String ret = null;
		ret = mtSrc.getProjectRelativePath().removeFirstSegments(1).removeLastSegments(1).toOSString().replace(File.separator, ".");
		return ret;
	}

}
