/*
 * Copyright (c) 2005-2008 Obeo
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Obeo - initial API and implementation
 */
package org.acceleo.gen.ui.wizards.export.model;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.runtime.IPath;
import org.eclipse.jdt.core.IClasspathEntry;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.JavaCore;
import org.eclipse.jdt.core.JavaModelException;

/**
 * Export utility class which contains general information about the standalone
 * project.
 * 
 * @author www.obeo.fr
 * 
 */
public class AcceleoExportModel {

	/** All plugins ID required by the standalone engine */
	public static final String[] SYSTEM_PLUGINS = {
			"org.acceleo.tools", "org.acceleo.template", "org.acceleo.engine", "org.acceleo.gen", "org.acceleo.parser", "org.eclipse.emf.common", "org.eclipse.emf.ecore", //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$ //$NON-NLS-6$ //$NON-NLS-7$ //$NON-NLS-8$
			"org.eclipse.emf.ecore.xmi", "org.eclipse.emf.mapping.ecore2xml", }; //$NON-NLS-1$ //$NON-NLS-2$

	/**
	 * Source project.
	 */
	private IJavaProject projectSource;

	/**
	 * Destination project.
	 */
	private IJavaProject projectDestination;

	/** List of templates (IFile) */
	private List srcMtFiles = new ArrayList();

	/** List of source folders (IPath) */
	private List srcFolders = new ArrayList();

	/** List of destination files (IFile) */
	private List srcDestFiles = new ArrayList();

	/** List of an EMT file information (AcceleoExportEMTModel) */
	private List emts = new ArrayList();

	/** List of JAR names (String) */
	private List jarFiles = new ArrayList();

	/** Set of ECORE models (IPath) */
	private Set ecoreModels = new HashSet();

	/** Indicates if we are on a UML model */
	private boolean umlModel;

	/** Indicates if we are on an ECORE model */
	private boolean ecoreModel;

	/**
	 * Constructor
	 * 
	 * @param source
	 *            source project to export
	 * @param destination
	 *            destination project
	 */
	public AcceleoExportModel(IJavaProject source, IJavaProject destination) {
		this.projectSource = source;
		this.projectDestination = destination;
	}

	/**
	 * Gets the templates (IFile).
	 * 
	 * @return a list of templates
	 */
	public List getSrcMtFiles() {
		return srcMtFiles;
	}

	/**
	 * Sets the templates (IFile).
	 * 
	 * @param mtFiles
	 *            are the templates
	 */
	public void setSrcMtFiles(List mtFiles) {
		this.srcMtFiles.addAll(mtFiles);
		// All mt files should be in a source folder
		for (Iterator itFile = mtFiles.iterator(); itFile.hasNext();) {
			IFile currentFile = (IFile) itFile.next();
			IPath projectPath = currentFile.getProjectRelativePath();
			if (!projectPath.isEmpty() && projectPath.segmentCount() > 0) {
				IPath srcPath = projectPath.removeLastSegments(projectPath.segmentCount() - 1);
				if (!srcFolders.contains(srcPath)) {
					srcFolders.add(srcPath);
				}
			}
		}
	}

	/**
	 * Gets the source folders (IPath).
	 * 
	 * @return the source folders
	 */
	public List getSrcFolders() {
		return srcFolders;
	}

	/**
	 * Compute class path entries.
	 * 
	 * @return class path entries
	 * @throws JavaModelException
	 */
	public IClasspathEntry[] getClassPathEntry() throws JavaModelException {
		// Copy initial class path entry
		List classpathEntries = new ArrayList();
		for (int i = 0; i < projectDestination.getRawClasspath().length; ++i) {
			classpathEntries.add(projectDestination.getRawClasspath()[i]);
		}
		// Add source folder wich contains .mt, .emt and Java
		for (Iterator itSrcFolder = getSrcFolders().iterator(); itSrcFolder.hasNext();) {
			IPath scrFolders = (IPath) itSrcFolder.next();
			IClasspathEntry srcEntry = JavaCore.newSourceEntry(projectDestination.getProject().getFullPath().append(scrFolders.segment(scrFolders.segmentCount() - 1)));
			if (!classpathEntries.contains(srcEntry)) {
				classpathEntries.add(srcEntry);
			}
		}
		// Add jar wich are in /lib
		for (Iterator itJarName = jarFiles.iterator(); itJarName.hasNext();) {
			String jarName = (String) itJarName.next();
			IClasspathEntry newEntry = JavaCore.newLibraryEntry(projectDestination.getProject().getLocation().append("lib").append(jarName), null, null);
			if (!classpathEntries.contains(newEntry)) {
				classpathEntries.add(newEntry);
			}
		}
		// Add jar of the source project
		IClasspathEntry newEntry = JavaCore.newLibraryEntry(projectDestination.getProject().getLocation().append("lib").append(projectSource.getProject().getName() + ".jar"), null, null);
		classpathEntries.add(newEntry);

		return (IClasspathEntry[]) classpathEntries.toArray(new IClasspathEntry[] {});
	}

	/**
	 * Gets destination files (IFile).
	 * 
	 * @return destination files
	 */
	public List getSrcDestFiles() {
		return srcDestFiles;
	}

	/**
	 * Gets a list of an EMT file information (AcceleoExportEMTModel).
	 * 
	 * @return EMT file information
	 */
	public List getEmts() {
		return emts;
	}

	/**
	 * Gets a list of JAR names (String).
	 * 
	 * @return a list of JAR names
	 */
	public List getJarFiles() {
		return jarFiles;
	}

	/**
	 * Indicates if we are on a UML model.
	 * 
	 * @return true if we are on a UML model
	 */
	public boolean isUmlModel() {
		return umlModel;
	}

	/**
	 * Indicates if we are on a UML model.
	 * 
	 * @param umlModel
	 *            must be true if we are on a UML model
	 */
	public void setUmlModel(boolean umlModel) {
		this.umlModel = umlModel;
	}

	/**
	 * Indicates if we are on an ECORE model.
	 * 
	 * @return true if we are on an ECORE model
	 */
	public boolean isEcoreModel() {
		return ecoreModel;
	}

	/**
	 * Indicates if we are on an ECORE model.
	 * 
	 * @param ecoreModel
	 *            must be true if we are on an ECORE model
	 */
	public void setEcoreModel(boolean ecoreModel) {
		this.ecoreModel = ecoreModel;
	}

	/**
	 * Gets ECORE models (IPath).
	 * 
	 * @return ECORE models
	 */
	public Set getEcoreModels() {
		return ecoreModels;
	}

}
