/*
 * Copyright (c) 2005-2008 Obeo
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Obeo - initial API and implementation
 */
package org.acceleo.gen.ui.wizards.export.operation;

import fr.obeo.acceleo.gen.template.TemplateSyntaxExceptions;
import fr.obeo.acceleo.gen.template.eval.ENodeException;
import fr.obeo.acceleo.tools.resources.Resources;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import org.acceleo.gen.ui.AcceleoGenUIMessages;
import org.acceleo.gen.ui.AcceleoGenUiPlugin;
import org.acceleo.gen.ui.wizards.export.model.AcceleoExportEMTModel;
import org.acceleo.gen.ui.wizards.export.model.AcceleoExportModel;
import org.acceleo.gen.ui.wizards.export.template.AcceleoAllMainClassModel;
import org.acceleo.gen.ui.wizards.export.template.AcceleoExportToStandaloneTemplateModel;
import org.acceleo.gen.ui.wizards.export.template.CreateMainJavaGenerator;
import org.acceleo.gen.ui.wizards.export.template.CreateReadmeGenerator;
import org.acceleo.gen.ui.wizards.export.util.AcceleoExportUtil;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IFolder;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IncrementalProjectBuilder;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.Assert;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.Path;
import org.eclipse.core.runtime.Platform;
import org.eclipse.jdt.core.IClasspathEntry;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.JavaCore;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.operation.IRunnableWithProgress;
import org.eclipse.ui.PlatformUI;
import org.osgi.framework.Bundle;

/**
 * Operation to export a project wich contains .mt in a standalone project. The
 * target project contains several java classes. Each java class contains a main
 * method that is able to launch a code generation in a standalone mode.
 * 
 * @author www.obeo.fr
 * 
 */
public class AcceleoExportToStandaloneOperation implements IRunnableWithProgress {

	/** Source project */
	private IJavaProject projectSource;

	/** Destination project */
	private IJavaProject projectDestination;

	/** Templates to export */
	private List mtFiles;

	/** Destination project name */
	private String projectDestName;

	/** Utility datas to realize the export */
	private AcceleoExportModel exportModel;

	/** Resources from other project to copy */
	private IPath[] resources;

	private AcceleoAllMainClassModel allMainClassModel;

	/**
	 * Constructor.
	 * 
	 * @param projectSource
	 *            is the source project
	 * @param mtFiles
	 *            are the templates to export
	 * @param projectDestName
	 *            is the destination project name
	 * @param resources
	 *            are the resources from other project to copy
	 */
	public AcceleoExportToStandaloneOperation(IProject projectSource, List mtFiles, String projectDestName, IPath[] resources) {
		this.projectSource = JavaCore.create(projectSource);
		Assert.isNotNull(projectSource);
		this.mtFiles = mtFiles;
		Assert.isNotNull(mtFiles);
		this.projectDestName = projectDestName;
		this.resources = resources;
	}

	/**
	 * Utility method to get and merge properties file
	 * 
	 * @return a map where the keys are the name of the property file and the
	 *         values the values of the property file
	 * @throws CoreException
	 * @throws IOException
	 */
	private Map getPropertiesFile() throws CoreException, IOException {
		// Get all properties file
		IFolder outputFolder = Resources.getOrCreateFolder(projectSource.getProject(), projectSource.getOutputLocation().removeFirstSegments(1), null);
		List propertiesFile = AcceleoExportUtil.listPropertiesFile(outputFolder);

		// Merge properties files with the same identifier
		Map sourceMapProperties = new HashMap();
		Map destMapProperties = new HashMap();
		for (Iterator itFile = propertiesFile.iterator(); itFile.hasNext();) {
			IFile file = (IFile) itFile.next();
			if (sourceMapProperties.get(file.getName()) == null) {
				sourceMapProperties.put(file.getName(), new ArrayList());
			}
			((List) (sourceMapProperties.get(file.getName()))).add(file);
		}
		List propertyFileNameMerged = new ArrayList();
		for (Iterator itName = sourceMapProperties.keySet().iterator(); itName.hasNext();) {
			String name = (String) itName.next();
			List currentPropertiesFile = (List) sourceMapProperties.get(name);
			if (currentPropertiesFile.size() > 0) {
				propertyFileNameMerged.add(name);
			}
			destMapProperties.put(name, AcceleoExportUtil.mergeProperties(currentPropertiesFile));
		}
		if (propertyFileNameMerged.size() > 0) {
			String files = "";
			for (int i = 0; i < propertyFileNameMerged.size(); ++i) {
				files += propertyFileNameMerged.get(i);
				if (i < propertyFileNameMerged.size() - 1) {
					files += ", ";
				}
			}
			MessageDialog.openWarning(PlatformUI.getWorkbench().getActiveWorkbenchWindow().getShell(), AcceleoGenUIMessages.getString("AcceleoExportToStandaloneOperation.PropertiesConflict"),
					AcceleoGenUIMessages.getString("AcceleoExportToStandaloneOperation.PropertiesNameConflict", new Object[] { files }));
		}
		return destMapProperties;
	}

	/* (non-Javadoc) */
	public void run(IProgressMonitor monitor) throws InvocationTargetException, InterruptedException {
		try {

			// Project creation
			projectDestination = Resources.createJavaProject(computeProjectName(), monitor);
			Assert.isNotNull(projectDestination);

			exportModel = new AcceleoExportModel(projectSource, projectDestination);
			allMainClassModel = new AcceleoAllMainClassModel();
			// find all .mt : except all include in output dir of the src
			// project
			exportModel.setSrcMtFiles(mtFiles);

			Map mapDestProperties = getPropertiesFile();

			for (Iterator itFile = exportModel.getSrcMtFiles().iterator(); itFile.hasNext();) {
				IFile file = (IFile) itFile.next();
				// Create directory in destination project
				Resources.getOrCreateFolder(projectDestination.getProject(), file.getProjectRelativePath().removeLastSegments(1), monitor);
				// Compile projectSource .mt => projectDestination .emt
				String emtDestFile = projectDestination.getProject().getLocation().append(file.getProjectRelativePath()).toFile().getAbsolutePath();
				emtDestFile = emtDestFile.substring(0, emtDestFile.length() - ".mt".length()) + ".emt";
				String currentNsURI = AcceleoExportUtil.compileTemplateFile(file, new File(emtDestFile));
				if (currentNsURI != null && currentNsURI.length() > 0) {
					AcceleoExportEMTModel exportTemplateModel = new AcceleoExportEMTModel();
					// check if we are working on a uml model
					if (currentNsURI.startsWith("http://www.eclipse.org/uml2/")) {
						exportModel.setUmlModel(true);
						exportTemplateModel.setUmlModel(true);
					}
					if (currentNsURI.startsWith("/") && currentNsURI.endsWith(".ecore")) {
						exportModel.setEcoreModel(true);
						exportModel.getEcoreModels().add(currentNsURI);
						exportTemplateModel.setEcoreModel(true);
					}
					exportTemplateModel.setNsURI(currentNsURI);
					String mtFilePath = file.getProjectRelativePath().removeFirstSegments(1).toPortableString();
					String emtFilePath = mtFilePath.substring(0, mtFilePath.length() - ".mt".length()) + ".emt";
					exportTemplateModel.setEmtFile(emtFilePath);
					exportTemplateModel.setMtSrc(file);
					exportModel.getEmts().add(exportTemplateModel);
				}

				// Copy properties file next the .emt file
				for (Iterator propertyFileNameIterator = mapDestProperties.keySet().iterator(); propertyFileNameIterator.hasNext();) {
					String propertyFileName = (String) propertyFileNameIterator.next();
					String propertyFile = projectDestination.getProject().getLocation().append(file.getProjectRelativePath().removeLastSegments(1)).append(propertyFileName).toFile().getAbsolutePath();
					FileOutputStream propertyOutputStream = new FileOutputStream(propertyFile);
					Properties properties = (Properties) mapDestProperties.get(propertyFileName);
					properties.store(propertyOutputStream, AcceleoGenUIMessages.getString("AcceleoExportToStandaloneOperation.GeneratedByAcceleoStandaloneExport"));
				}
			}

			// Create lib directory
			IFolder libFolder = Resources.getOrCreateFolder(projectDestination.getProject(), projectDestination.getProject().getProjectRelativePath().append("lib"), monitor);

			// Add uml jar
			String umlJarPath = "";
			String umlResourceJarPath = "";
			String umlCommonJarPath = "";
			// If we have one or more template wich use an uml model
			if (exportModel.isUmlModel()) {
				// Get org.eclipse.uml2.uml
				umlJarPath = AcceleoExportUtil.exportPlugin(libFolder.getLocation().toFile(), Platform.getBundle("org.eclipse.uml2.uml"));
				// Get org.eclipse.uml2.uml.resources
				umlResourceJarPath = AcceleoExportUtil.exportPlugin(libFolder.getLocation().toFile(), Platform.getBundle("org.eclipse.uml2.uml.resources"));
				// Get org.eclipse.uml2.uml.resources
				umlCommonJarPath = AcceleoExportUtil.exportPlugin(libFolder.getLocation().toFile(), Platform.getBundle("org.eclipse.uml2.common"));
			}

			// Generate for each .emt a Java class (use Jet)
			CreateMainJavaGenerator mainJavaGenerator = new CreateMainJavaGenerator();
			for (Iterator itExportTemplateModel = exportModel.getEmts().iterator(); itExportTemplateModel.hasNext();) {
				AcceleoExportEMTModel exportTemplateModel = (AcceleoExportEMTModel) itExportTemplateModel.next();
				AcceleoExportToStandaloneTemplateModel templateModel = new AcceleoExportToStandaloneTemplateModel();
				templateModel.setEmtFile(exportTemplateModel.getEmtFile());
				templateModel.setEmtName(exportTemplateModel.getEmtName());
				templateModel.setModelPackageName(exportTemplateModel.getModelPackageName());
				templateModel.setPackageName(exportTemplateModel.getPackageName());
				allMainClassModel.getClasses().add(templateModel.getPackageName() + "." + templateModel.getEmtName());
				// We have to add more informations when we are on a uml model
				if (exportTemplateModel.isUmlModel()) {
					templateModel.setUmlModel(true);
					// TODO refactore with dependencies list
					templateModel.setUmlJarPath((libFolder.getLocation().toFile().getAbsolutePath() + File.separator + umlJarPath).replace(File.separator, "/"));
					templateModel.setUmlResourcesJarPath((libFolder.getLocation().toFile().getAbsolutePath() + File.separator + umlResourceJarPath).replace(File.separator, "/"));
					templateModel.setUmlCommonJarPath((libFolder.getLocation().toFile().getAbsolutePath() + File.separator + umlCommonJarPath).replace(File.separator, "/"));
				} else if (exportTemplateModel.isEcoreModel()) {
					templateModel.setEcoreModel(true);
					templateModel.setEcorePath(exportTemplateModel.getNsURI());
					// update scr folders
					if (!exportModel.getSrcFolders().contains(new Path("resources"))) {
						exportModel.getSrcFolders().add(new Path("resources"));
					}
					// Copy modele to destination project
					IFolder folder = Resources.getOrCreateFolder(projectDestination.getProject(), new Path("resources").append(exportTemplateModel.getNsURI()).removeLastSegments(1), monitor);
					File source = Resources.findResource(new Path(exportTemplateModel.getNsURI())).getLocation().toFile();
					File destination = new File(folder.getLocation().toOSString() + File.separator + new Path(exportTemplateModel.getNsURI()).lastSegment());
					AcceleoExportUtil.copyFile(source, destination);
				}
				Resources.createFile(projectDestination.getProject(), exportTemplateModel.getMtSrc().getProjectRelativePath().removeLastSegments(1).append(exportTemplateModel.getEmtName() + ".java"),
						mainJavaGenerator.generate(templateModel), monitor);
			}

			// For each nsURI, add the corresponding jar in lib folder
			for (Iterator itEmtModel = exportModel.getEmts().iterator(); itEmtModel.hasNext();) {
				AcceleoExportEMTModel emtModel = (AcceleoExportEMTModel) itEmtModel.next();
				// If we are not working on an ecore or an uml model
				if (!emtModel.getNsURI().startsWith("/") && !emtModel.getNsURI().startsWith("http://www.eclipse.org/uml2/")) {
					String jarFile = AcceleoExportUtil.exportMetamodels(libFolder.getLocation().toFile(), emtModel.getNsURI());
					exportModel.getJarFiles().add(jarFile);
				}
			}

			// For each system dependencies, add the corresponding jar in lib
			// folder
			for (int i = 0; i < AcceleoExportModel.SYSTEM_PLUGINS.length; ++i) {
				Bundle plugin = Platform.getBundle(AcceleoExportModel.SYSTEM_PLUGINS[i]);
				String jarFile = AcceleoExportUtil.exportPlugin(libFolder.getLocation().toFile(), plugin);
				exportModel.getJarFiles().add(jarFile);
			}
			// Jar the source project in destination
			// project/lib/projectSourceName.jar
			AcceleoExportUtil.exportProject(libFolder.getLocation().toFile(), projectSource);

			// Resolve source project class path
			IClasspathEntry[] classPathEntries = projectSource.getResolvedClasspath(true);
			List systemClassPathEntries = AcceleoExportUtil.getSystemClassPathEntry(projectSource);
			List javaProjects = new ArrayList();
			for (int i = 0; i < classPathEntries.length; ++i) {
				// We don't want system libraries like jre
				if (!systemClassPathEntries.contains(classPathEntries[i])) {
					// if we are on a library we need to copy it in lib folder
					if (classPathEntries[i].getEntryKind() == IClasspathEntry.CPE_LIBRARY) {
						String jarName = classPathEntries[i].getPath().lastSegment();
						// We check that we are not on a system plugin
						boolean systemPlugin = false;
						for (int s = 0; s < exportModel.SYSTEM_PLUGINS.length && !systemPlugin; ++s) {
							systemPlugin = jarName.startsWith(exportModel.SYSTEM_PLUGINS[s]);
						}
						if (!systemPlugin) {
							exportModel.getJarFiles().add(jarName);
							AcceleoExportUtil.copyFile(classPathEntries[i].getPath().toFile(), libFolder.getLocation().append(jarName).toFile());
						}
					}
					// if we are on a project we need to jar it and copy it in
					// lib folder
					else if (classPathEntries[i].getEntryKind() == IClasspathEntry.CPE_PROJECT) {
						IProject project = ResourcesPlugin.getWorkspace().getRoot().getProject(classPathEntries[i].getPath().lastSegment());
						IJavaProject javaProject = JavaCore.create(project);
						javaProjects.add(javaProject);
					}
					// if we are on a variable
					else if (classPathEntries[i].getEntryKind() == IClasspathEntry.CPE_VARIABLE) {
						// TODO
					}
				}
			}

			// Export all JavaProject dependencies
			while (javaProjects.size() > 0) {
				IJavaProject currentJavaProject = (IJavaProject) javaProjects.remove(0);
				// We ensure that the current project has not been already
				// exported
				if (!exportModel.getJarFiles().contains(currentJavaProject.getProject().getName() + ".jar")) {
					AcceleoExportUtil.exportProject(libFolder.getLocation().toFile(), currentJavaProject);
					exportModel.getJarFiles().add(currentJavaProject.getProject().getName() + ".jar");
					IClasspathEntry[] currentClassPathEntries = currentJavaProject.getResolvedClasspath(true);
					for (int i = 0; i < currentClassPathEntries.length; ++i) {
						if (currentClassPathEntries[i].getEntryKind() == IClasspathEntry.CPE_PROJECT) {
							IProject project = ResourcesPlugin.getWorkspace().getRoot().getProject(currentClassPathEntries[i].getPath().lastSegment());
							IJavaProject javaProject = JavaCore.create(project);
							javaProjects.add(javaProject);
						}
					}
				}
			}

			// If we work on uml model we must have some jar in the classpath
			if (exportModel.isUmlModel()) {
				if (!exportModel.getJarFiles().contains(umlJarPath)) {
					exportModel.getJarFiles().add(umlJarPath);
				}
				if (!exportModel.getJarFiles().contains(umlResourceJarPath)) {
					exportModel.getJarFiles().add(umlResourceJarPath);
				}
				if (!exportModel.getJarFiles().contains(umlCommonJarPath)) {
					exportModel.getJarFiles().add(umlCommonJarPath);
				}
			}

			// Copy all resource needed
			copyResources(monitor);

			// Create a readme.txt
			Resources.appendFile(projectDestination.getProject(), new Path("Readme.txt"), new CreateReadmeGenerator().generate(null), monitor);

			// TODO improve the jet generator to manage multi metamodels
			// projects and drope leading point when templates aren't in a
			// pachage
			// Create a Generator.java wich call all main classes
			// Resources.appendFile(projectDestination.getProject(), new
			// Path("src").append(new Path("Generator.java")), new
			// CreateMainAllJavaGenerator().generate(allMainClassModel),
			// monitor);

			// Update the source folders and lib folder
			projectDestination.setRawClasspath(exportModel.getClassPathEntry(), monitor);

			// Force refresh and build of the destination project
			projectDestination.getProject().refreshLocal(IResource.DEPTH_INFINITE, monitor);
			projectDestination.getProject().build(IncrementalProjectBuilder.INCREMENTAL_BUILD, monitor);

		} catch (CoreException e) {
			AcceleoGenUiPlugin.getDefault().log(e, true);
		} catch (TemplateSyntaxExceptions e) {
			AcceleoGenUiPlugin.getDefault().log(e, true);
		} catch (ENodeException e) {
			AcceleoGenUiPlugin.getDefault().log(e, true);
		} catch (IOException e) {
			AcceleoGenUiPlugin.getDefault().log(e, true);
		}
	}

	/**
	 * Utility method which copy resources.
	 * 
	 * @param monitor
	 *            is the monitor
	 * @throws IOException
	 * @throws CoreException
	 */
	private void copyResources(IProgressMonitor monitor) throws IOException, CoreException {
		if (resources != null && resources.length > 0) {
			// update scr folders
			if (!exportModel.getSrcFolders().contains(new Path("resources"))) {
				exportModel.getSrcFolders().add(new Path("resources"));
			}
			for (int i = 0; i < resources.length; ++i) {
				IPath path = resources[i];
				if (path.segmentCount() > 1) {

					// Copy modele to destination project
					IFolder folder = Resources.getOrCreateFolder(projectDestination.getProject(), new Path("resources").append(path).removeLastSegments(1), monitor);
					File source = Resources.findResource(path).getLocation().toFile();
					File destination = new File(folder.getLocation().toOSString() + File.separator + path.lastSegment());
					AcceleoExportUtil.copyFile(source, destination);
				}
			}
		}
	}

	/**
	 * Compute the project name.
	 * 
	 * @return the project name
	 */
	private String computeProjectName() {
		String ret = projectDestName;
		if (ret == null || ret.length() == 0) {
			ret = projectSource.getProject().getName() + ".export";
		}
		return ret;
	}

}
