/*
 * Copyright (c) 2005-2008 Obeo
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Obeo - initial API and implementation
 */
package org.acceleo.gen.ui.wizards.export.util;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.net.URL;
import java.nio.channels.FileChannel;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;
import java.util.jar.Attributes;
import java.util.jar.JarEntry;
import java.util.jar.JarOutputStream;
import java.util.jar.Manifest;

import org.acceleo.gen.ui.AcceleoGenUiPlugin;
import org.acceleo.gen.ui.builders.AcceleoCompiler;
import org.eclipse.core.resources.IContainer;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.FileLocator;
import org.eclipse.core.runtime.Platform;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.plugin.EcorePlugin;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.resource.impl.ResourceSetImpl;
import org.eclipse.jdt.core.IClasspathContainer;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.JavaCore;
import org.eclipse.jdt.core.JavaModelException;
import org.osgi.framework.Bundle;

import fr.obeo.acceleo.gen.template.TemplateSyntaxExceptions;
import fr.obeo.acceleo.gen.template.eval.ENodeException;
import fr.obeo.acceleo.gen.template.scripts.SpecificScript;

/**
 * Utility class for the AcceleoExportStandaloneOperation operation.
 * 
 * @author www.obeo.fr
 * 
 */
public class AcceleoExportUtil {

	/**
	 * Gets all the templates in the given container and its children.
	 * 
	 * @param container
	 *            is the container
	 * @param exceptContainer
	 *            is a container to ignore
	 * @return all the templates
	 * @throws CoreException
	 */
	public static List listTemplateFiles(IContainer container, IContainer exceptContainer) throws CoreException {
		List res = new ArrayList();
		IResource[] children = container.members();
		if (children != null) {
			for (int i = 0; i < children.length; ++i) {
				if (children[i] instanceof IContainer) {
					if (exceptContainer == null || !children[i].equals(exceptContainer)) {
						res.addAll(listTemplateFiles((IContainer) children[i], exceptContainer));
					}
				} else if (children[i] instanceof IFile) {
					if (children[i].getName().endsWith(".mt")) { //$NON-NLS-1$
						res.add((IFile) children[i]);
					}
				}
			}
		}
		return res;
	}

	/**
	 * Gets all the properties files in the given container and its children.
	 * 
	 * @param container
	 *            is the container
	 * @return all the properties files
	 * @throws CoreException
	 */
	public static List listPropertiesFile(IContainer container) throws CoreException {
		List res = new ArrayList();
		IResource[] children = container.members();
		if (children != null) {
			for (int i = 0; i < children.length; ++i) {
				if (children[i] instanceof IContainer) {
					res.addAll(listPropertiesFile((IContainer) children[i]));
				} else if (children[i] instanceof IFile) {
					if (children[i].getName().endsWith(".properties")) {
						res.add((IFile) children[i]);
					}
				}
			}
		}
		return res;
	}

	/**
	 * Properties files amalgamation.
	 * 
	 * @param propertiesFile
	 *            are the files to merge
	 * @return the merging result
	 * @throws IOException
	 * @throws FileNotFoundException
	 */
	public static Properties mergeProperties(List propertiesFile) throws IOException {
		Properties ret = new Properties();
		for (Iterator it = propertiesFile.iterator(); it.hasNext();) {
			IFile file = (IFile) it.next();
			Properties currentProperties = new Properties();
			currentProperties.load(new FileInputStream(file.getLocation().toFile()));
			Enumeration keys = currentProperties.keys();
			while (keys.hasMoreElements()) {
				Object currentKey = keys.nextElement();
				ret.put(currentKey, currentProperties.get(currentKey));
			}
		}
		return ret;
	}

	/**
	 * Compiles the template file and returns the metamodel URI.
	 * 
	 * @param templateFile
	 *            is the template file
	 * @param dest
	 *            is the EMT file to create
	 * @return the URI in use in the template
	 * @throws TemplateSyntaxExceptions
	 * @throws ENodeException
	 * @throws IOException
	 */
	public static String compileTemplateFile(IFile templateFile, File dest) throws TemplateSyntaxExceptions, ENodeException, IOException {
		String res = null;
		SpecificScript template = new SpecificScript(templateFile.getLocation().toFile(), null, null);
		template.reset();
		if (template.hasFileTemplate()) {
			AcceleoCompiler compiler = new AcceleoCompiler();
			org.acceleo.template.ResourceSet emt;
			emt = compiler.export(template);
			if (emt != null) {
				ResourceSet resourceSet = new ResourceSetImpl();
				String emtFileName = dest.getAbsolutePath();

				Resource resource = resourceSet.createResource(URI.createFileURI(emtFileName));
				resource.getContents().add(emt);
				Iterator it = emt.getResources().iterator();
				while (res == null && it.hasNext()) {
					org.acceleo.template.Resource r = (org.acceleo.template.Resource) it.next();
					if (r instanceof org.acceleo.template.core.Metamodel) {
						res = ((org.acceleo.template.core.Metamodel) r).getName();
					}
				}
				resource.save(null);
				resource.unload();
			}
		}
		return res;
	}

	/**
	 * Export a jar containing the definition of the meta model.
	 * 
	 * @param outputDirectory
	 *            is the output directory
	 * @param nsURI
	 *            is the URI of the metamodel to export
	 * @return the jar name
	 */
	public static String exportMetamodels(File outputDirectory, String nsURI) {
		String res = null;
		EPackage ePkg = (EPackage) EPackage.Registry.INSTANCE.get(nsURI);
		if (ePkg != null) {
			URI pluginURI = (URI) EcorePlugin.getEPackageNsURIToGenModelLocationMap().get(nsURI);

			String[] segments = pluginURI.segments();
			if ("plugin".equals(segments[0])) { //$NON-NLS-1$
				Bundle plugin = Platform.getBundle(segments[1]);
				res = exportPlugin(outputDirectory, plugin);
			} else if ("platform".equals(segments[0])) { //$NON-NLS-1$
				// ResourcesPlugin.getWorkspace();
			}
		}
		return res;
	}

	/**
	 * Get recursively all accessible files and directories from the given file.
	 * 
	 * @param in
	 *            is the input file
	 * @return all accessible files and directories
	 */
	private static List listFiles(File in) {
		List res = new ArrayList();
		File[] children = in.listFiles();
		if (children != null) {
			for (int i = 0; i < children.length; ++i) {
				if (children[i].isDirectory()) {
					if (!".svn".equals(children[i].getName())) {
						res.addAll(listFiles(children[i]));
					}
				} else {
					res.add((File) children[i]);
				}
			}
		}
		return res;
	}

	/**
	 * Creates a Jar from a JavaProject in the given folder (the first
	 * argument). The name of the jar is based on the project name.
	 * 
	 * @param out
	 *            is the folder where to put the new jar file.
	 * @param javaProject
	 *            is the project to build
	 */
	public static void exportProject(File out, IJavaProject javaProject) {
		// TODO generate the correct manifest
		try {
			if (out.isDirectory()) {
				String basePath = javaProject.getProject().getFolder(javaProject.getOutputLocation().removeFirstSegments(1)).getLocation().toFile().getAbsolutePath();
				String jarFileName = javaProject.getProject().getName() + ".jar";
				JarOutputStream outputStream = new JarOutputStream(new FileOutputStream(out.getAbsolutePath() + File.separator + jarFileName));
				jarFiles(basePath, listFiles(new File(basePath)), outputStream);
				outputStream.closeEntry();
				outputStream.close();
			}
		} catch (CoreException e) {
			AcceleoGenUiPlugin.getDefault().log(e, true);
		} catch (IOException e) {
			AcceleoGenUiPlugin.getDefault().log(e, true);
		}
	}

	/**
	 * Creates a Jar from a plugin in the given folder (the first argument). The
	 * name of the jar is based on the plugin name.
	 * 
	 * @param out
	 *            is the folder where to put the new jar file.
	 * @param plugin
	 *            is the plugin to build
	 * @return the jar name
	 */
	public static String exportPlugin(File out, Bundle plugin) {
		String res = null;
		try {
			URL location = FileLocator.resolve(plugin.getEntry("/")); //$NON-NLS-1$
			// We already have a jar : we have to copy it a the right place
			if ("jar".equals(location.getProtocol())) { //$NON-NLS-1$
				File jarFile = new File(location.getPath().substring("file:".length(), location.getPath().length() - "!/".length())); //$NON-NLS-1$ //$NON-NLS-2$
				copyFile(jarFile, new File(out.getAbsolutePath() + File.separator + jarFile.getName()));
				res = jarFile.getName();
				// other we have to package it in a jar file
			} else if ("file".equals(location.getProtocol())) { //$NON-NLS-1$
				File manifestFile = new File(location.getPath() + "META-INF" + File.separator + "MANIFEST.MF"); //$NON-NLS-1$ //$NON-NLS-2$
				Manifest manifest = new Manifest(new FileInputStream(manifestFile));
				if (manifestFile.isFile()) {
					manifest = new Manifest(new FileInputStream(manifestFile));
				} else {
					manifest = new Manifest();
				}
				if (manifest.getMainAttributes().get(new Attributes.Name("Manifest-Version")) == null) { //$NON-NLS-1$
					manifest.getMainAttributes().put(new Attributes.Name("Manifest-Version"), "1.0"); //$NON-NLS-1$ //$NON-NLS-2$
				}
				if (manifest.getMainAttributes().get(new Attributes.Name("Created-By")) == null) { //$NON-NLS-1$
					manifest.getMainAttributes().put(new Attributes.Name("Created-By"), "Acceleo exporter"); //$NON-NLS-1$ //$NON-NLS-2$
				}
				File jarFile = new File(out.getAbsolutePath() + File.separator + plugin.getSymbolicName() + ".jar"); //$NON-NLS-1$
				res = plugin.getSymbolicName() + ".jar"; //$NON-NLS-1$
				JarOutputStream jar = new JarOutputStream(new FileOutputStream(jarFile), manifest);

				try {
					// ./bin
					File in = new File(location.getPath() + "bin" + File.separator); //$NON-NLS-1$
					List toJar = listFiles(in);
					jarFiles(in.getAbsolutePath(), toJar, jar);

					// ./model/*
					in = new File(location.getPath() + "model" + File.separator); //$NON-NLS-1$
					toJar = listFiles(in);
					// plugin.properties
					File ppFile = new File(location.getPath() + File.separator + "plugin.properties"); //$NON-NLS-1$
					if (ppFile.exists()) {
						toJar.add(ppFile);
					}
					// plugin.xml
					File pxFile = new File(location.getPath() + File.separator + "plugin.xml"); //$NON-NLS-1$
					if (pxFile.exists()) {
						toJar.add(pxFile);
					}
					// about.html
					File ahFile = new File(location.getPath() + File.separator + "about.html"); //$NON-NLS-1$
					if (ahFile.exists()) {
						toJar.add(ahFile);
					}

					jarFiles(location.getPath(), toJar, jar);
				} finally {
					jar.closeEntry();
					jar.close();
				}
			}
		} catch (IOException e) {
			AcceleoGenUiPlugin.getDefault().log(e, true);
		}
		return res;
	}

	/**
	 * Copy file from source to destination.
	 * 
	 * @param in
	 *            is the input file
	 * @param out
	 *            is the output file
	 * @throws IOException
	 */
	public static void copyFile(File in, File out) throws IOException {
		FileChannel inChannel = new FileInputStream(in).getChannel();
		FileChannel outChannel = new FileOutputStream(out).getChannel();
		try {
			inChannel.transferTo(0, inChannel.size(), outChannel);
		} catch (IOException e) {
			throw e;
		} finally {
			if (inChannel != null)
				inChannel.close();
			if (outChannel != null)
				outChannel.close();
		}
	}

	/**
	 * Exports a list of file in a single jar.
	 * 
	 * @param basePath
	 *            is the base directory of the files to export
	 * @param inputFiles
	 *            are the files to export
	 * @param out
	 *            is the jar output stream
	 * @throws IOException
	 */
	private static void jarFiles(String basePath, List inputFiles, JarOutputStream out) throws IOException {
		final byte[] buffer = new byte[100 * 1024];
		int length;

		Iterator it = inputFiles.iterator();
		while (it.hasNext()) {
			File file = (File) it.next();
			if (file.getAbsolutePath().startsWith(basePath)) {
				String entryName = file.getAbsolutePath().substring(basePath.length());
				if (entryName.startsWith(File.separator)) {
					entryName = entryName.substring(1);
				}
				JarEntry entry = new JarEntry(entryName.replace(File.separator, "/"));
				out.putNextEntry(entry);
				FileInputStream fileStream = new FileInputStream(file);

				while ((length = fileStream.read(buffer)) >= 0) {
					out.write(buffer, 0, length);
				}
				fileStream.close();
			}
		}
	}

	/**
	 * Computes system libraries which are in the class path.
	 * 
	 * @param javaProject
	 *            is the java project
	 * @return class path entries
	 * @throws JavaModelException
	 */
	public static List getSystemClassPathEntry(IJavaProject javaProject) throws JavaModelException {
		List ret = new ArrayList();
		for (int i = 0; i < javaProject.getRawClasspath().length; ++i) {
			IClasspathContainer classPathContainer = JavaCore.getClasspathContainer(javaProject.getRawClasspath()[i].getPath(), javaProject);
			if (classPathContainer != null && (classPathContainer.getKind() == IClasspathContainer.K_DEFAULT_SYSTEM || classPathContainer.getKind() == IClasspathContainer.K_SYSTEM)) {
				for (int j = 0; j < classPathContainer.getClasspathEntries().length; ++j) {
					ret.add(classPathContainer.getClasspathEntries()[j]);
				}
			}
		}
		return ret;
	}
}
