/*
 * Copyright (c) 2008 Obeo.
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Obeo - initial API and implementation
 */
package org.acceleo.engine;

import fr.obeo.acceleo.gen.template.eval.ENode;
import fr.obeo.acceleo.gen.template.eval.ENodeCastException;
import fr.obeo.acceleo.gen.template.eval.ENodeException;

import org.acceleo.template.ResourceSet;
import org.acceleo.template.TemplatePackage;
import org.acceleo.template.core.Metamodel;
import org.acceleo.template.core.Script;

import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.acceleo.engine.context.EvalContext;
import org.acceleo.engine.template.impl.spec.IEvaluable;
import org.acceleo.engine.template.impl.spec.TResourceSet;
import org.acceleo.tools.factory.FactoryException;
import org.acceleo.tools.log.EvalLog;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EObject;

/**
 * Utility class for EMT model handling.
 * 
 * @author <a href="mailto:david.bonneau@nomail.none">David Bonneau</a>
 * @author www.obeo.fr
 * 
 */
public class EMT {

	/** Resources (Templates, Metamodel, Services) used during the generation. */
	private ResourceSet resourceSet;

	/** Generation source metamodel. */
	private Metamodel metamodel;

	/** Script execution context. */
	private EvalContext evalContext;

	/** Storage for generation logs. */
	private EvalLog logs = new EvalLog();

	/**
	 * Constructor.
	 * 
	 * @param resourceSet
	 *            context ResourceSet
	 * @param file
	 *            is the XMI representation of EMT model
	 */
	public EMT(org.eclipse.emf.ecore.resource.ResourceSet resourceSet,
			URI templateURI, ClassLoader loader) {
		// Register the EMT metamodel.
		TemplatePackage.eINSTANCE.eClass();

		// TODO Refactored this part
		File file = new File(templateURI.path());
		Map<String, List<File>> propertyContainer = new HashMap<String, List<File>>();
		List<File> currentList = new ArrayList<File>();
		currentList.add(file.getParentFile());
		propertyContainer.put(STANDALONE_EXECUTION_CONTEXT, currentList);

		// Demand load the resource for this file.
		this.resourceSet = (ResourceSet) (resourceSet.getResource(templateURI,
				true).getContents().get(0));
		evalContext = new EvalContext(propertyContainer, this.resourceSet,
				loader);
	}

	public final static String STANDALONE_EXECUTION_CONTEXT = "STANDALONE_EXECUTION_CONTEXT";

	/**
	 * Constructor.
	 * 
	 * @param templateModel
	 *            in-memory EMT
	 */
	public EMT(ResourceSet templateModel,
			Map<String, List<File>> propertyContainer, ClassLoader loader) {
		// Register the EMT metamodel
		TemplatePackage.eINSTANCE.eClass();

		this.resourceSet = templateModel;

		evalContext = new EvalContext(propertyContainer, this.resourceSet,
				loader);
	}

	/**
	 * Returns metamodel of the source model.
	 * 
	 * @return metamodel of source model
	 */
	public Metamodel getMetamodel() {
		if (metamodel == null) {
			for (int j = 0; j < resourceSet.getResources().size()
					&& metamodel == null; ++j) {
				if (resourceSet.getResources().get(j) instanceof Metamodel) {
					metamodel = (Metamodel) resourceSet.getResources().get(j);
				}
			}
		}
		return metamodel;
	}

	/**
	 * Generate all file names and their content.
	 * 
	 * @param model
	 *            the model to generate
	 * @param recursive
	 *            is recursive or not
	 * @return a map filename -> generated content
	 * @throws ENodeCastException
	 *             if generation failed
	 * @throws ENodeException
	 *             if generation failed
	 * @throws FactoryException
	 *             if the EMT metamodel's specific implementation is not
	 *             reachable if the metamodel of the EMT and the model aren't
	 *             the same if generation failed
	 */
	// TODO use FileContentMap ?
	// TODO manage Acceleo 2.x IGenFilter
	public Map<String, String> gen(IModel model, boolean recursive)
			throws FactoryException, ENodeException, ENodeCastException {
		Map<String, String> ret = new HashMap<String, String>();
		// Checking of specific code access.
		if (metamodelSpecificImplFound()) {
			// Checking of metamodel compatibility.
			// TODO notify using a warning is nsURI are different
			// if (getMetamodel().getName().equals(model.getNsURI())) {
			for (EObject root : model.getRoots()) {
				ret.putAll(genString(root));
				if (recursive) {
					Iterator<EObject> it = root.eAllContents();
					while (it.hasNext()) {
						ret.putAll(genString(it.next()));
					}
				}
			}
			// } else {
			// throw new
			// FactoryException(AcceleoEngineMessages.getString("EMT.IncompatibleMetamodel",
			// new Object[] { getMetamodel().getName(), model.getNsURI(), }));
			// //$NON-NLS-1$
			// }
		} else {
			throw new FactoryException(AcceleoEngineMessages
					.getString("EMT.MetamodelSpecificImplNotFound")); //$NON-NLS-1$
		}
		return ret;
	}

	/**
	 * Tell if we have access to the specific implementation of the EMT
	 * metamodel.
	 * 
	 * @return true if the EMT metamodel is reachable
	 */
	public boolean metamodelSpecificImplFound() {
		return resourceSet instanceof TResourceSet;
	}

	/**
	 * Generate files name and files content.
	 * 
	 * @param root
	 *            the EObject to use for the generation
	 * @param recursive
	 *            should we explore objects contained in the root object?
	 * @return the mapping beetwen files name and file content
	 * @throws ENodeCastException
	 *             if generation failed
	 * @throws ENodeException
	 *             if generation failed
	 * @throws FactoryException
	 *             if generation failed
	 */
	public Map<String, String> gen(EObject root, boolean recursive)
			throws FactoryException, ENodeException, ENodeCastException {
		Map<String, String> ret = new HashMap<String, String>();

		ret.putAll(genString(root));
		if (recursive) {
			Iterator<EObject> it = root.eAllContents();
			while (it.hasNext()) {
				ret.putAll(genString(it.next()));
			}
		}

		return ret;
	}

	/**
	 * Generate a list of file name that can be generated from the given
	 * EObject.
	 * 
	 * @param root
	 *            the EObject to generate
	 * @return the file name list
	 * @throws ENodeException
	 *             if the generation of the file name fail.
	 * @throws FactoryException
	 *             if the generation of the file name fail. if metamodel
	 *             specific implementation can't be found.
	 * @throws ENodeCastException
	 *             if the generation of the file name fail. if the generated
	 *             file name can't be casted to a string.
	 */
	public List<String> getFileNames(EObject root) throws FactoryException,
			ENodeException {
		List<String> res = new LinkedList<String>();
		Map<String, ENode> generated = genENodes(root, false);

		res.addAll(generated.keySet());
		return res;
	}

	/**
	 * Generate a map with file name as key and file content as value.
	 * 
	 * @param root
	 *            the EObject to generate
	 * @return the file name to content mapping
	 * @throws ENodeException
	 *             if the generation of the file name fail.
	 * @throws FactoryException
	 *             if the generation of the file name fail. if metamodel
	 *             specific implementation can't be found.
	 * @throws ENodeCastException
	 *             if the generation of the file name fail. if the generated
	 *             file name can't be casted to a string.
	 */
	public Map<String, String> genString(EObject root) throws FactoryException,
			ENodeException, ENodeCastException {
		Map<String, String> res = new HashMap<String, String>();
		Map<String, ENode> generated = genENodes(root);

		for (Entry<String, ENode> entry : generated.entrySet()) {
			res.put(entry.getKey(), entry.getValue().asString());
		}
		return res;
	}

	/**
	 * Generate a map with file name as key and the generated ENode as value.
	 * 
	 * @param root
	 *            the EObject to generate
	 * @return the file name to ENode mapping
	 * @throws ENodeException
	 *             if the generation of the file name fail.
	 * @throws FactoryException
	 *             if the generation of the file name fail. if metamodel
	 *             specific implementation can't be found.
	 * @throws ENodeCastException
	 *             if the generation of the file name fail. if the generated
	 *             file name can't be casted to a string.
	 */
	public Map<String, ENode> genENodes(EObject root) throws FactoryException,
			ENodeException {
		return genENodes(root, true);
	}

	/**
	 * Generate a map with file name as key and the generated ENode as value.
	 * 
	 * @param root
	 *            the EObject to generate
	 * @param genContent
	 *            generate content or not
	 * @return the file name to ENode mapping
	 * @throws ENodeException
	 *             if the generation of the file name fail.
	 * @throws FactoryException
	 *             if the generation of the file name fail. if metamodel
	 *             specific implementation can't be found.
	 */
	private Map<String, ENode> genENodes(EObject root, boolean genContent)
			throws FactoryException, ENodeException {
		if (metamodelSpecificImplFound()) {
			Map<String, ENode> ret = new HashMap<String, ENode>();
			// Checking of metamodel compatibility.
			List<Script> scripts = evalContext.getApplicableScripts(root);
			for (Script script : scripts) {
				evalContext.startScript(script);
				// A script that generates files is found.
				if (script.getDescriptor().getFile() != null) {
					ENode fileName = ((IEvaluable) script.getDescriptor()
							.getFile()).gen(new ENode(root, (IEvaluable) script
							.getDescriptor().getFile(), evalContext
							.getLaunchManager().isSynchronize()), evalContext);
					// File generation.
					ENode node = null;
					if (genContent) {
						node = ((IEvaluable) script).gen(new ENode(root,
								(IEvaluable) script, evalContext
										.getLaunchManager().isSynchronize()),
								evalContext);
					} else {
						node = new ENode(ENode.EMPTY, null);
					}

					logs.getAll(node.log(), false);

					if (!"".equals(fileName.asString())) { //$NON-NLS-1$
						ret.put(fileName.asString().trim(), node);
					}
					break; // generate only the first script with file template
				}
				evalContext.endScript();
			}
			return ret;
		} else {
			throw new FactoryException(AcceleoEngineMessages
					.getString("EMT.MetamodelSpecificImplNotFound")); //$NON-NLS-1$
		}
	}

	/**
	 * Getter for the current EMT resource set.
	 * 
	 * @return the current resource set
	 */
	public ResourceSet getResourceSet() {
		return resourceSet;
	}

	/**
	 * Getter for the logger.
	 * 
	 * @return the current logger
	 */
	public EvalLog getLogs() {
		return logs;
	}

	public void resetLogs() {
		this.logs = new EvalLog();
	}
}
