/*
 * Copyright (c) 2008 Obeo.
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Obeo - initial API and implementation
 */
package org.acceleo.engine.context;

import fr.obeo.acceleo.gen.template.eval.ILauncherManager;
import fr.obeo.acceleo.gen.template.eval.LaunchManager;

import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Stack;

import org.acceleo.engine.eval.imports.EvalJavaService;
import org.acceleo.engine.template.impl.spec.IEvaluable;
import org.acceleo.parser.template.TemplateSyntaxException;
import org.acceleo.parser.template.expressions.ExpressionParser;
import org.acceleo.template.Resource;
import org.acceleo.template.ResourceSet;
import org.acceleo.template.core.Script;
import org.acceleo.template.core.ScriptDescriptor;
import org.acceleo.template.core.Template;
import org.acceleo.tools.factory.ETools;
import org.acceleo.tools.strings.Int2;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EObject;

/**
 * Eval context.
 * 
 * @author David Bonneau
 * 
 */
public class EvalContext {
	/**
	 * Mapping between EClass and scripts that can be applied on objects of this
	 * type.
	 */
	private Map<EClass, List<Script>> compatibleScripts = new HashMap<EClass, List<Script>>();

	/** Stack for the script contexts. */
	private Stack<ScriptContext> scriptContexts = new Stack<ScriptContext>();

	/** Stack for the template contexts. */
	private Stack<TemplateContext> templateContexts = new Stack<TemplateContext>();

	/** List templates. */
	private List<Template> templates;

	/** The current file used by this context. */
	private Map<String, List<File>> propertyContainer;

	/** Mapping between a template and its template context. */
	private Map<Template, TemplateContext> templateContextFactory = new HashMap<Template, TemplateContext>();

	/** The current resourceSet (EMT elements). */
	private ResourceSet resourceSet;

	/** Launch manager */
	ILauncherManager launchManager;

	/** ClassLoader for Java services */
	private ClassLoader loader;

	/**
	 * Constructor.
	 * 
	 * @param file
	 *            the file used with this context.
	 * @param resourceSet
	 *            the EMT resource set
	 */
	public EvalContext(Map<String, List<File>> propertyContainer, ResourceSet resourceSet, ClassLoader loader) {
		this.propertyContainer = propertyContainer;
		this.resourceSet = resourceSet;
		this.loader = loader;
	}

	/**
	 * Getter for the template context of the given template.
	 * 
	 * @param template
	 *            the template from witch we want the template context
	 * @return the template context if exists, null otherwise
	 */
	public TemplateContext getTemplateContext(Template template) {
		if (templateContextFactory.get(template) == null) {
			templateContextFactory.put(template, new TemplateContext(template, loader));
		}
		return templateContextFactory.get(template);
	}

	/**
	 * Start a new script context for the given Script.
	 * 
	 * @param script
	 *            the Script to start
	 */
	public void startScript(Script script) {
		// TODO g�rer les erreurs si on est encore sur le m�me script
		// TODO get properties container from script
		ScriptContext currentContext = new ScriptContext(script, propertyContainer, this);
		TemplateContext templateContext = getTemplateContext(script.getTemplate());
		templateContext.initConstants();

		scriptContexts.push(currentContext);
		templateContexts.push(templateContext);
	}

	/**
	 * End the last started Script.
	 */
	public void endScript() {
		// TODO Manage error if there is an empty context
		scriptContexts.pop();
		templateContexts.pop();

		if (!templateContexts.isEmpty()) {
			templateContexts.peek().initConstants();
		}
	}

	/**
	 * Returns current script context.
	 * 
	 * @return current script context
	 */
	public ScriptContext getCurrentScriptContext() {
		// TODO Manage error if there is an empty context
		return scriptContexts.peek();
	}

	/**
	 * Getter for the current template context.
	 * 
	 * @return the current template context.
	 */
	public TemplateContext getCurrentTemplateContext() {
		return templateContexts.peek();
	}

	/**
	 * Get the Script with a given name that can be applied to the given
	 * EClassifier.
	 * 
	 * @param eClass
	 *            the EClassifier on witch the Script can be applied
	 * @param name
	 *            the name of the script we are looking for
	 * @return the matching Script if founded, null otherwise
	 */
	public Script getScript(EObject eObj, String name) {
		Script script = null;
		// search in specifics templates
		for (int i = 0; i < (templateContexts.size() - 1) && script == null; ++i) {
			script = templateContexts.get(i).getScript(eObj, name);
		}

		// search in current template
		if (script == null) {
			script = getCurrentTemplateContext().getScript(eObj, name);
		}

		// search in imported templates
		if (script == null) {
			Iterator<Resource> it = getCurrentTemplateContext().getTemplate().getImports().iterator();
			while (script == null && it.hasNext()) {
				Resource imp = it.next();
				if (imp instanceof Template) {
					script = getTemplateContext((Template) imp).getScript(eObj, name);
				}
			}
		}

		return script;
	}

	/**
	 * Getter for the currently usable services.
	 * 
	 * @return the list of currently usable services
	 */
	public List<EvalJavaService> getCurrentServices() {
		List<EvalJavaService> ret = new ArrayList<EvalJavaService>();
		ret.addAll(getCurrentTemplateContext().getImportServices());
		ret.addAll(getCurrentScriptContext().getServices());
		return ret;
	}

	/**
	 * Computes available templates.
	 * 
	 * @return a map with the available templates
	 */
	public List<Template> getTemplates() {
		if (templates == null) {
			templates = new ArrayList<Template>();
			for (int j = 0; j < resourceSet.getResources().size(); ++j) {
				if (resourceSet.getResources().get(j) instanceof Template) {
					templates.add((Template) resourceSet.getResources().get(j));
				}
			}
		}
		return templates;
	}

	/**
	 * Returns the list of applicable scripts for the object.
	 * 
	 * @param object
	 *            current eObject for which applicable scripts will be found.
	 * @return the list of applicable scripts
	 */
	public List<Script> getApplicableScripts(EObject object) {
		List<Script> ret = new ArrayList<Script>();
		List<EClass> compat = new ArrayList<EClass>();

		// let's get all compatible types
		compat.add(object.eClass());
		compat.addAll(object.eClass().getEAllSuperTypes());
		for (EClass cls : compat) {
			ret.addAll(subGetApplicableScripts(cls));
		}
		return ret;
	}

	/**
	 * Sub-method of getApplicableScripts(EObject object). It returns the list
	 * of applicable scripts for the current EClass.
	 * 
	 * @param eClass
	 *            current eClass for which applicable scripts will be found.
	 * @return the list of applicable scripts for the current EClass.
	 */
	private List<Script> subGetApplicableScripts(EClass eClass) {
		List<Script> ret = compatibleScripts.get(eClass);

		if (ret == null) {
			ret = new ArrayList<Script>();
			String type;

			// Calcul du type par rapport aux information du object
			type = ETools.getEClassifierPath(eClass);
			// Recherche des scripts qui vont bien dans les templates
			for (Template template : getTemplates()) {
				for (Script script : template.getScripts()) {
					if (script.getDescriptor() != null) {
						ScriptDescriptor descriptor = script.getDescriptor();
						if (descriptor.getType().equals(type) || type.endsWith("." + descriptor.getType())) { //$NON-NLS-1$
							ret.add(script);
						}
					}
				}
			}
			compatibleScripts.put(eClass, ret);
		}
		return ret;
	}

	/**
	 * Compilation of a string representing an Acceleo expression.
	 * 
	 * @param offset
	 *            current ENode
	 * @param expression
	 *            the string representing the expression
	 * @return result of the compilation
	 * @throws TemplateSyntaxException
	 *             If the expression is not valid.
	 */
	public IEvaluable compile(int offset, String expression) throws TemplateSyntaxException {
		return (IEvaluable) ExpressionParser.fromString(offset, expression, new Int2(0, expression.length()), null);
	}

	public ILauncherManager getLaunchManager() {
		if (launchManager == null) {
			launchManager = LaunchManager.create("run", false); //$NON-NLS-1$
		}
		return launchManager;
	}

	public void setLaunchManager(ILauncherManager launchManager) {
		this.launchManager = launchManager;
	}
}
