/*
 * Copyright (c) 2008 Obeo.
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Obeo - initial API and implementation
 */
package org.acceleo.engine.context;

import fr.obeo.acceleo.gen.template.eval.ENode;

import org.acceleo.template.core.Script;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Stack;

import org.acceleo.engine.eval.imports.EvalJavaService;
import org.acceleo.engine.eval.imports.services.ENodeServices;
import org.acceleo.engine.eval.imports.services.PropertiesServices;
import org.acceleo.engine.eval.imports.services.RequestServices;

/**
 * Script evaluation context.
 * 
 * @author David Bonneau
 * @author www.obeo.fr
 * 
 */
public class ScriptContext {

	/** Context key for while index. */
	public static final Integer WHILE_INDEX = new Integer(0);

	/** Context key for template arguments. */
	public static final Integer TEMPLATE_ARGS = new Integer(1);

	/** Context key for service position. */
	public static final Integer ARGUMENT_POSITION = new Integer(2);

	/** Context key for current node. */
	public static final Integer CURRENT_NODE = new Integer(3);

	/** Context key for current node. */
	public static final Integer TEMPLATE_NODE = new Integer(4);

	/** System services. */
	private static List<EvalJavaService> systemServices;

	/**
	 * Context that is used during generation.
	 */
	private Map<Object, Stack<Object>> context = new HashMap<Object, Stack<Object>>();

	/** Current script. */
	private Script script;

	// /** System resource (.emt) containing the script. */
	// private File file;

	/**
	 * Utility class for properties management.
	 */
	private ScriptProperties properties;

	/**
	 * Constructor.
	 * 
	 * @param script
	 *            the script to use in this context
	 * @param file
	 *            the file that contains the script
	 * @param evalContext
	 *            the evaluation context to use
	 */
	public ScriptContext(Script script, Map propertyContainer, EvalContext evalContext) {
		this.script = script;
		properties = new ScriptProperties(script, propertyContainer);
		systemServices = new ArrayList<EvalJavaService>();
		systemServices.add(new EvalJavaService(new ENodeServices(evalContext), true));
		systemServices.add(new EvalJavaService(new RequestServices(evalContext), true));
		systemServices.add(new EvalJavaService(new PropertiesServices(this, evalContext), true));
	}

	public Script getScript() {
		return script;
	}

	/**
	 * Returns the value to which the context of this generator maps the
	 * specified key.
	 * 
	 * @param key
	 *            whose associated value is to be returned
	 * @return the value to which the context of this generator maps the
	 *         specified key
	 */
	public Object contextPeek(Object key) {
		Object res = null;
		Stack<Object> stack = context.get(key);
		if (stack != null && !stack.isEmpty()) {
			Object result = stack.peek();
			if (result instanceof ENode) {
				res = ((ENode) result).copy();
			} else if (result instanceof ENode[]) {
				ENode[] copy = new ENode[((ENode[]) result).length];
				for (int i = 0; i < copy.length; i++) {
					copy[i] = ((ENode[]) result)[i].copy();
				}
				res = copy;
			} else {
				res = result;
			}
		} else {
			res = null;
		}
		return res;
	}

	/**
	 * Returns the value to which the context of this generator maps the
	 * specified key at the given index.
	 * 
	 * @param key
	 *            is the key
	 * @param index
	 *            is the index
	 * @return the value to which the context of this generator maps the
	 *         specified key
	 */
	public Object contextAt(Object key, int index) {
		Object res = null;
		Stack<Object> stack = context.get(key);
		if (stack != null) {
			if (index >= 0 && index < stack.size()) {
				Object result = stack.elementAt(stack.size() - index - 1);
				if (result instanceof ENode) {
					res = ((ENode) result).copy();
				} else if (result instanceof ENode[]) {
					ENode[] copy = new ENode[((ENode[]) result).length];
					for (int i = 0; i < copy.length; i++) {
						copy[i] = ((ENode[]) result)[i].copy();
					}
					res = copy;
				} else {
					res = result;
				}
			} else {
				res = null;
			}
		} else {
			res = null;
		}
		return res;
	}

	/**
	 * Associates the specified value with the specified key in the context of
	 * this generator.
	 * 
	 * @param key
	 *            with which the specified value is to be associated
	 * @param value
	 *            to be associated with the specified key
	 */
	public void contextPush(Object key, Object value) {
		Stack<Object> stack = context.get(key);
		if (stack == null) {
			stack = new Stack<Object>();
			context.put(key, stack);
		}
		stack.push(value);
	}

	/**
	 * Deletes the last specified value for the specified key in the context of
	 * this generator.
	 * 
	 * @param key
	 *            is the key
	 */
	public void contextPop(Object key) {
		Stack<Object> stack = context.get(key);
		if (stack != null) {
			stack.pop();
		}
	}

	/**
	 * Getter for properties.
	 * 
	 * @return properties attribute
	 */
	public ScriptProperties getProperties() {
		return properties;
	}

	/**
	 * Getter for systemServices.
	 * 
	 * @return systemServices attribute
	 */
	public List<EvalJavaService> getServices() {
		return systemServices;
	}

	// public File getFile() {
	// return file;
	// }
}
