/*
 * Copyright (c) 2008 Obeo.
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Obeo - initial API and implementation
 */
package org.acceleo.engine.context;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.acceleo.engine.eval.imports.EvalJavaService;
import org.acceleo.engine.eval.imports.JavaServiceNotFoundException;
import org.acceleo.engine.eval.imports.services.ContextServices;
import org.acceleo.engine.eval.imports.services.EObjectServices;
import org.acceleo.engine.eval.imports.services.ResourceServices;
import org.acceleo.engine.eval.imports.services.StringServices;
import org.acceleo.engine.eval.imports.services.XpathServices;
import org.acceleo.template.Resource;
import org.acceleo.template.core.Metamodel;
import org.acceleo.template.core.Script;
import org.acceleo.template.core.Service;
import org.acceleo.template.core.Template;
import org.acceleo.tools.factory.ETools;
import org.acceleo.tools.template.TemplateConstants;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.EcorePackage;

/**
 * Current template management.
 * 
 * @author David Bonneau
 * @author www.obeo.fr
 * 
 */
public class TemplateContext {
	/** Map script name -> script. */
	private Map<String, Script> scripts;

	/** List of system services. */
	private static List<EvalJavaService> systemServices;

	/** Current template. */
	private Template template;

	/** List of accessible services. */
	private List<EvalJavaService> javaServices;

	/**
	 * Constructor.
	 * 
	 * @param template
	 *            is the current template
	 */
	public TemplateContext(Template template, ClassLoader loader) {
		this.template = template;

		javaServices = new ArrayList<EvalJavaService>();
		javaServices.addAll(getSystemServices());
		// Initialisation des imports de services
		for (Resource currentImport : template.getImports()) {
			if (currentImport instanceof Service) {
				Service currentService = (Service) currentImport;
				try {
					EvalJavaService eval = new EvalJavaService(currentService.getName(), loader);
					javaServices.add(eval);
				} catch (JavaServiceNotFoundException e) {
					// TODO exception - throw the exception when old services
					// form Acceleo 2.x will be removed from the EMT
					// (fr.obeo.acceleo...)
				}
			}
		}
	}

	/**
	 * System services management.
	 * 
	 * @return the list of system services
	 */
	private List<EvalJavaService> getSystemServices() {
		// TODO voir l'intérêt des SystemServices avec le StandAlone (.emt)
		// ils ne sont plus implicites ! dans les imports ils sont toujours
		// présents
		// on doit donc les rajoutés en double avec ce mécanisme, à confirmer
		// !
		if (systemServices == null) {
			try {
				systemServices = new ArrayList<EvalJavaService>();
				// TODO que devient les notions de root, normalement cela
				// devrait fonctionner ?
				// TODO que devient la notion de specificScript
				systemServices.add(new EvalJavaService(StringServices.class.getName(), StringServices.class.getClassLoader()));
				systemServices.add(new EvalJavaService(EObjectServices.class.getName(), EObjectServices.class.getClassLoader()));
				systemServices.add(new EvalJavaService(XpathServices.class.getName(), XpathServices.class.getClassLoader()));
				systemServices.add(new EvalJavaService(ResourceServices.class.getName(), ResourceServices.class.getClassLoader()));
				systemServices.add(new EvalJavaService(ContextServices.class.getName(), ContextServices.class.getClassLoader()));
			} catch (JavaServiceNotFoundException e) {
				// TODO je ne sais pas trop comment gérer cela
				systemServices = null;
			}
		}
		return systemServices;
	}

	public List<EvalJavaService> getImportServices() {
		return javaServices;
	}

	/**
	 * Initializes all the constants for the generation tool.
	 */
	public void initConstants() {
		TemplateConstants.initConstants(template.getBeginTag() + template.getEndTag());
	}

	/**
	 * Get the Script with a given name that can be applied to the given
	 * EObject.
	 * 
	 * @param eObj
	 *            the source object of the script
	 * @param name
	 *            the name of the script
	 * @return the script
	 */
	public Script getScript(EObject eObj, String name) {
		EClass eClass = eObj.eClass();
		Script script = getScript(eClass, name);
		if (script == null) {
			if (eClass.getESuperTypes().isEmpty() && eClass != EcorePackage.eINSTANCE.getEObject()) {
				/*
				 * We want to add an "artificial" specialization on EObject
				 */
				script = getScript(EcorePackage.eINSTANCE.getEObject(), name);
			} else {
				script = getScriptFromSupertype(eClass.getESuperTypes(), name);
			}
		}
		return script;
	}

	/**
	 * Gets script from supertype if applicable.
	 * 
	 * @param superTypes
	 *            on which searched script could be available
	 * @param name
	 *            script name
	 * @return script from supertype if applicable
	 */
	private Script getScriptFromSupertype(EList<EClass> superTypes, String name) {
		EClass superType = null;
		Script script = null;
		for (Iterator<EClass> it = superTypes.iterator(); it.hasNext() && script == null;) {
			superType = it.next();
			script = getScript(superType, name);
		}
		if (script == null) {
			for (Iterator<EClass> it = superTypes.iterator(); it.hasNext() && script == null;) {
				script = getScriptFromSupertype(it.next().getESuperTypes(), name);
			}
		}
		if (script != null) {
			scripts.put(script.getDescriptor().getName() + ETools.getEClassifierShortPath(superType), script);
		}
		return script;
	}

	/**
	 * Get the Script with a given name that can be applied to the given
	 * EClassifier in the current template only (imports not scanned).
	 * 
	 * @param eClass
	 *            the EClassifier on witch the Script can be applied
	 * @param name
	 *            the name of the script we are looking for
	 * @return the matching Script if founded, null otherwise
	 */
	private Script getScript(EClassifier eClass, String name) {
		if (scripts == null) {
			scripts = new HashMap<String, Script>();
			String nsURI = null;
			for (Resource resource : template.getImports()) {
				if (resource instanceof Metamodel) {
					nsURI = ((Metamodel) resource).getName();
				}
			}
			EPackage metamodel = EPackage.Registry.INSTANCE.getEPackage(nsURI);
			if (metamodel == null) {
				metamodel = ETools.ecore2EPackage(nsURI, true);
			}
			// scan Scripts of the current template
			for (Script script : template.getScripts()) {
				if (script.getDescriptor() != null) {
					EClassifier scriptType = ETools.getEClassifier(metamodel, script.getDescriptor().getType());
					if (scriptType != null) {
						scripts.put(script.getDescriptor().getName() + ETools.getEClassifierShortPath(scriptType), script);
					}
				}
			}
		}
		return scripts.get(name + ETools.getEClassifierShortPath(eClass));
	}

	public Template getTemplate() {
		return template;
	}
}
