/*
 * Copyright (c) 2005-2008 Obeo.
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Obeo - initial API and implementation
 */

package org.acceleo.engine.eval.imports;

import fr.obeo.acceleo.gen.template.eval.ENode;
import fr.obeo.acceleo.gen.template.eval.ENodeCastException;
import fr.obeo.acceleo.gen.template.eval.ENodeException;
import fr.obeo.acceleo.gen.template.eval.ENodeList;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.TreeSet;

import org.acceleo.engine.AcceleoEngineMessages;
import org.acceleo.engine.context.EvalContext;
import org.acceleo.engine.context.ScriptContext;
import org.acceleo.engine.eval.imports.services.ContextServices;
import org.acceleo.engine.eval.imports.services.ENodeServices;
import org.acceleo.engine.eval.imports.services.EObjectServices;
import org.acceleo.engine.eval.imports.services.RequestServices;
import org.acceleo.template.expressions.Call;
import org.acceleo.template.expressions.Expression;
import org.acceleo.template.expressions.StringLiteral;
import org.acceleo.tools.factory.ETools;
import org.acceleo.tools.factory.FactoryException;
import org.acceleo.tools.strings.Int2;
import org.acceleo.tools.template.TemplateConstants;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.EcorePackage;

/**
 * Java services element that can be used during a code generation.
 * 
 * @author www.obeo.fr
 * 
 */
public class EvalJavaService {

	/**
	 * The default evaluation mode.
	 */
	public static final int MODE_DEFAULT = 0;

	/**
	 * The 'ENode' evaluation mode : ENode services are available.
	 */
	public static final int MODE_ENODE = 1;

	/**
	 * The 'ENodeList' evaluation mode : ENodeList services are available.
	 */
	public static final int MODE_LIST = 2;

	/**
	 * Select keyword.
	 */
	private static final String SELECT = "select"; //$NON-NLS-1$

	/**
	 * Keywords: services with type resolve.
	 */
	private static final String[] SERVICES_WITH_TYPE_RESOLVE = new String[] { "filter", "cast", "until", "eContainer", "eAllContents", "current", }; //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$ //$NON-NLS-6$

	/**
	 * Keywords: services with type bridge.
	 */
	private static final String[] SERVICES_WITH_TYPE_BRIDGE = new String[] { SELECT, "delete", "sep", "nGet", "trace", "debug", "nFirst", "nLast", "sort", "nSort", "minimize", "nMinimize", "reverse", //$NON-NLS-1$ //$NON-NLS-2$ //$NON-NLS-3$ //$NON-NLS-4$ //$NON-NLS-5$ //$NON-NLS-6$ //$NON-NLS-7$ //$NON-NLS-8$ //$NON-NLS-9$ //$NON-NLS-10$ //$NON-NLS-11$ //$NON-NLS-12$
			"nReverse", }; //$NON-NLS-1$

	/**
	 * List of type bridge.
	 */
	private static List<String> typeBridgeList;

	/**
	 * List of type resolve.
	 */
	private static List<String> typeResolveList;

	/**
	 * The evaluation mode.
	 */
	protected int mode = MODE_DEFAULT;

	/**
	 * Instance that contains the services.
	 */
	protected Object instance;

	/**
	 * The class loader.
	 */
	private ClassLoader loader;

	/**
	 * Gets the service for the given name.
	 */
	private Map<String, Method[]> name2service = new HashMap<String, Method[]>();

	/**
	 * Cache for type resolving.
	 */
	private Map<List<Object>, Object> quickResolveType = new HashMap<List<Object>, Object>();

	/**
	 * Indicates if the services have a script context.
	 */
	private boolean hasScriptContext;

	/**
	 * Constructor.
	 * 
	 * @param name
	 *            is the class to be created
	 * @throws JavaServiceNotFoundException
	 *             if the class instance can't be created
	 */
	public EvalJavaService(String name, ClassLoader loader) throws JavaServiceNotFoundException {
		this.loader = loader;
		instance = instanciate(name, this.loader);
		hasScriptContext = false;
		initializeName2service();
	}

	/**
	 * Constructor.
	 * 
	 * @param instance
	 *            Instance
	 * @param hasScriptContext
	 */
	public EvalJavaService(Object instance, boolean hasScriptContext) {
		this.instance = instance;
		this.loader = instance.getClass().getClassLoader();
		this.hasScriptContext = hasScriptContext;
		initializeName2service();
	}

	/**
	 * Instanciate the service with the given name.
	 * 
	 * @param name
	 *            name of the service
	 * @return the instance of the service
	 * @throws JavaServiceNotFoundException
	 *             if the service can't be found
	 */
	private Object instanciate(String name, ClassLoader loader) throws JavaServiceNotFoundException {
		Object res = null;
		String className = name;

		// Remark : script "fr.package.template" -> class "fr.package.Template"
		String shortName = className;
		String path = ""; //$NON-NLS-1$
		int jDot = className.lastIndexOf("."); //$NON-NLS-1$
		if (jDot > -1) {
			shortName = className.substring(jDot + 1);
			path = className.substring(0, jDot);
		}
		if (shortName.length() > 0)
			shortName = shortName.substring(0, 1).toUpperCase() + shortName.substring(1);
		className = ((path.length() > 0) ? path + '.' : "") + shortName; //$NON-NLS-1$

		try {
			res = loader.loadClass(className).newInstance();
		} catch (InstantiationException e) {
			throw new JavaServiceNotFoundException(e.getMessage());
		} catch (IllegalAccessException e) {
			throw new JavaServiceNotFoundException(e.getMessage());
		} catch (ClassNotFoundException e) {
			throw new JavaServiceNotFoundException(e.getMessage());
		}
		return res;
	}

	/**
	 * Creates (if necessary) and returns list of type resolve.
	 * 
	 * @return list of type resolve
	 */
	private static List<String> getTypeResolveList() {
		if (typeResolveList == null) {
			typeResolveList = new ArrayList<String>();
			for (int i = 0; i < SERVICES_WITH_TYPE_RESOLVE.length; i++) {
				typeResolveList.add(SERVICES_WITH_TYPE_RESOLVE[i]);
			}
		}
		return typeResolveList;
	}

	/**
	 * Creates (if necessary) and returns list of type bridge.
	 * 
	 * @return list of type bridge
	 */
	private static List<String> getTypeBridgeList() {
		if (typeBridgeList == null) {
			typeBridgeList = new ArrayList<String>();
			for (int i = 0; i < SERVICES_WITH_TYPE_BRIDGE.length; i++) {
				typeBridgeList.add(SERVICES_WITH_TYPE_BRIDGE[i]);
			}
		}
		return typeBridgeList;
	}

	public int getMode() {
		return mode;
	}

	public void setMode(int mode) {
		this.mode = mode;
	}

	/**
	 * Returns true if the services have a script context.
	 * 
	 * @return true if the services have a script context
	 */
	public boolean hasScriptContext() {
		return hasScriptContext;
	}

	public Object getInstance() {
		return instance;
	}

	/**
	 * Initialize the mapping for direct access to service's methods with the
	 * name .
	 */
	private void initializeName2service() {
		name2service.clear();
		if (instance != null) {
			Method[] methods = instance.getClass().getMethods();
			for (int i = 0; i < methods.length; i++) {
				if (methods[i].getDeclaringClass() == instance.getClass() && methods[i].getParameterTypes().length > 0) {
					String key = methods[i].getName() + methods[i].getParameterTypes().length;
					Method[] name2values = name2service.get(key);
					if (name2values == null) {
						name2service.put(key, new Method[] { methods[i] });
					} else {
						List<Method> nameToValues = new ArrayList<Method>(name2values.length + 1);
						boolean ok = false;
						for (int j = 0; j < name2values.length; j++) {
							if (name2values[j].getParameterTypes()[0].isAssignableFrom(methods[i].getParameterTypes()[0]) || name2values[j].getParameterTypes()[0] == ENode.class) {
								nameToValues.add(methods[i]);
								ok = true;
							}
							nameToValues.add(name2values[j]);
						}
						if (!ok) {
							nameToValues.add(methods[i]);
						}
						name2service.put(key, nameToValues.toArray(new Method[nameToValues.size()]));
					}
				}
			}
		}
	}

	/**
	 * Gives an array of potential service's methods match according to the
	 * given name and number of arguments expected by the service.
	 * 
	 * @param name
	 *            the name of the service
	 * @param argsCount
	 *            the number of arguments expected by the service
	 * @return an array with potential service's methods
	 */
	private Method[] getPotentialMethods(String name, int argsCount) {
		String key = name + argsCount;
		return name2service.get(key);
	}

	/**
	 * Checks if the call is applicable on the current JavaService.
	 * 
	 * @param call
	 *            the current Call
	 * @param args
	 *            arguments of the current Call
	 * @return the value generated by the service or null if service is not
	 *         found
	 */
	public boolean isApplicable(Call call, ENode[] args) {
		// TODO Not used for the moment, check if it could be used to filter
		// find service method
		if (name2service.containsKey(call.getName() + (args.length + 1))) {
			return true;
		} else {
			return false;
		}
	}

	/**
	 * Generate the Call using a Java service.
	 * 
	 * @param call
	 *            the current Call
	 * @param node
	 *            the current ENode
	 * @param evalContext
	 *            the current evalutation context
	 * @param args
	 *            arguments of the current Call
	 * @return the value generated by the service or null if service is not
	 *         found
	 * @throws FactoryException
	 *             if call of the service's method failed
	 * @throws ENodeException
	 *             if call of the service's method failed
	 */
	public ENode gen(Call call, ENode node, EvalContext evalContext, ENode[] args) throws FactoryException, ENodeException {
		if (isApplicable(call, args)) {
			return genSub(call, node, evalContext, args);
		} else {
			return null;
		}
	}

	/**
	 * Generate the Call using a Java service.
	 * 
	 * @param call
	 *            the current Call
	 * @param node
	 *            the current ENode
	 * @param evalContext
	 *            the current evalutation context
	 * @param args
	 *            arguments of the current Call
	 * @return the value generated by the service or null if service is not
	 *         found
	 * @throws FactoryException
	 *             if call of the service's method failed
	 * @throws ENodeException
	 *             if call of the service's method failed
	 */
	private ENode genSub(Call call, ENode node, EvalContext evalContext, ENode[] args) throws FactoryException, ENodeException {
		// ASSERT (instance != null){
		// Get the type and the value for each argument
		List<Class<?>> argTypesList = new ArrayList<Class<?>>(args.length + 1);
		List<Object> argValuesList = new ArrayList<Object>(args.length + 1);
		argTypesList.add(null);
		argValuesList.add(null);
		for (int i = 0; i < args.length; i++) {
			ENode arg = args[i];
			argTypesList.add(arg.getTypeClass());
			argValuesList.add(arg.getValue());
		}
		Class<?>[] argTypes = argTypesList.toArray(new Class<?>[argTypesList.size()]);
		Object[] argValues = argValuesList.toArray();
		// Get the method
		Method m = null;
		if (mode == MODE_DEFAULT && !node.isNull()) {
			argTypes[0] = node.getTypeClass();
			argValues[0] = node.getValue();
			m = eGetMethod(call.getName(), argTypes, argValues[0]);
		} else if (mode == MODE_ENODE) {
			argTypes[0] = ENode.class;
			argValues[0] = node;
			m = eGetMethod(call.getName(), argTypes, argValues[0]);
		} else if (mode == MODE_LIST && node.isList()) {
			argTypes[0] = ENodeList.class;
			if (node.isList()) {
				try {
					argValues[0] = node.getList();
					m = eGetMethod(call.getName(), argTypes, argValues[0]);
				} catch (ENodeCastException e1) {
					m = null;
				}
			} else {
				m = null;
			}
			if (m == null) {
				try {
					argValues[0] = node.getList().asList();
					m = eGetMethod(call.getName(), argTypes, argValues[0]);
				} catch (ENodeCastException e2) {
					m = null;
				}
			}
		}
		if (m == null && instance != null) {
			// Try to adapt parameters
			m = getMethodWithAdaptedValues(call, node, args, argTypes, argValues);
		}
		if (m != null) {
			// Invoke method on instance
			return tryInvokeOnInstance(call, node, evalContext, m, argValues);
		} else {
			// Service doesn't exist
			return null;
		}
	}

	/**
	 * Compute the string to display in the log for the given method.
	 * 
	 * @param m
	 *            the methode to display in the log
	 * @return the string to display
	 */
	private String getDisplayString(Method m) {
		// Display method
		String displayString = m.getName() + " ("; //$NON-NLS-1$
		Class<?>[] paramTypes = m.getParameterTypes();
		for (int j = 1; j < paramTypes.length; j++) { // The first
			// parameter is
			// ignored
			Class<?> paramType = paramTypes[j];
			displayString += getSimpleName(paramType);
			if (j + 1 < paramTypes.length) {
				displayString += ", "; //$NON-NLS-1$
			}
		}
		displayString += ')';
		if (m.getReturnType() != null) {
			displayString += ' ' + getSimpleName(m.getReturnType());
		}
		displayString += " - " + getSimpleName(m.getDeclaringClass()); //$NON-NLS-1$
		return displayString;
	}

	/**
	 * Gets the simple name of the given class.
	 * <p>
	 * Sample : return 'Object' for 'java.lang.Object'
	 * 
	 * @param c
	 *            is the class
	 * @return the name
	 */
	public static String getSimpleName(Class<?> c) {
		String name = c.getName();
		int i = name.lastIndexOf("."); //$NON-NLS-1$
		if (i > -1) {
			name = name.substring(i + 1);
		}
		return name;
	}

	/**
	 * Adapt parameters values to match a method.
	 * 
	 * @param call
	 *            the current Call
	 * @param node
	 *            the current ENode
	 * @param args
	 *            arguments of the current Call
	 * @param argTypes
	 *            return adapted types
	 * @param argValues
	 *            return adapted values
	 * @return the method
	 */
	private Method getMethodWithAdaptedValues(Call call, ENode node, ENode[] args, Class<?>[] argTypes, Object[] argValues) {
		Method m = null;
		Method[] methods = getPotentialMethods(call.getName(), argTypes.length);
		if (methods != null) {
			for (int i = 0; m == null && i < methods.length; i++) {
				boolean ok = true;
				Class<?>[] parameterTypes = methods[i].getParameterTypes();
				if (parameterTypes.length > 0) {
					ok = mode == MODE_DEFAULT && adaptParameterType(parameterTypes[0], node, 0, argTypes, argValues);
					boolean isENode = mode == MODE_ENODE && parameterTypes[0] == ENode.class;
					boolean isList = mode == MODE_LIST && parameterTypes[0] == ENodeList.class;
					if (ok || isENode || isList) {
						ok = true;
						for (int j = 1; j < parameterTypes.length && ok; j++) {
							ok = adaptParameterType(parameterTypes[j], args[j - 1], j, argTypes, argValues);
						}
						if (ok) {
							m = methods[i];
						}
					}
				}
			}
		}
		return m;
	}

	/**
	 * Adapt the value to the type parameterType and store it at pos position.
	 * 
	 * @param parameterType
	 *            the type to adapt to
	 * @param value
	 *            the value to adapt
	 * @param pos
	 *            the position in the argument array for storing the result
	 * @param argTypes
	 *            array of parameters types
	 * @param argValues
	 *            array of parameters values
	 * @return true if type was adapted
	 */
	private boolean adaptParameterType(Class<?> parameterType, ENode value, int pos, Class<?>[] argTypes, Object[] argValues) {
		boolean ok = true;
		Class<?> adaptedType = ENode.getAdapterType(parameterType);
		if (adaptedType != null) {
			argTypes[pos] = adaptedType;
			try {
				argValues[pos] = value.getAdapterValue(adaptedType);
			} catch (ENodeCastException e) {
				ok = false;
			}
		} else {
			ok = false;
		}
		return ok;
	}

	/**
	 * Try to invoke the method and return the new corresponding ENode.
	 * 
	 * @param call
	 *            the current call to generate
	 * @param node
	 *            the current ENode
	 * @param evalContext
	 *            the current evaluation context
	 * @param m
	 *            the method to invoke
	 * @param argValues
	 *            argument values
	 * @return a new ENode containing the result of the method call
	 * @throws ENodeException
	 *             if a java runtime exception was thrown if an
	 *             InvocationTargetException that contain a ENodeException was
	 *             thrown
	 * @throws FactoryException
	 *             if an InvocationTargetException that contain a
	 *             FactoryException was thrown
	 */
	private ENode tryInvokeOnInstance(Call call, ENode node, EvalContext evalContext, Method m, Object[] argValues) throws ENodeException, FactoryException {
		Object result = null;
		if (SELECT.equals(call.getName()) && call.getArguments().size() > 0) {
			Int2 pos = new Int2(call.getArguments().get(0).getBegin(), call.getArguments().get(0).getEnd());
			evalContext.getCurrentScriptContext().contextPush(ScriptContext.ARGUMENT_POSITION, pos);
		}
		try {
			// void => ""
			if (m.getReturnType() != null && "void".equals(getSimpleName(m.getReturnType()))) { //$NON-NLS-1$
				m.invoke(instance, argValues);
				result = ""; //$NON-NLS-1$
			} else {
				result = m.invoke(instance, argValues);
			}
		} catch (IllegalArgumentException e1) {
			if (mode == MODE_DEFAULT && argValues.length > 0 && argValues[0] == null) {
				result = null;
			} else {
				throw e1;
			}
		} catch (IllegalAccessException e) {
			if (mode == MODE_DEFAULT && argValues.length > 0 && argValues[0] == null) {
				result = null;
			} else {
				final StringBuffer errorMessage = new StringBuffer("\n"); //$NON-NLS-1$
				errorMessage.append(e.getClass().getName());
				errorMessage.append(" : "); //$NON-NLS-1$
				errorMessage.append(e.getMessage());
				throw new ENodeException(
						AcceleoEngineMessages.getString("EvalJavaService.RuntimeException", new Object[] { getDisplayString(m), errorMessage.toString(), }), new Int2(call.getBegin(), call //$NON-NLS-1$
								.getEnd()), call.getTemplate(), node, true, e);
			}
		} catch (InvocationTargetException e) {
			if (mode == MODE_DEFAULT && argValues.length > 0 && argValues[0] == null) {
				result = null;
			} else {
				if (e.getTargetException() instanceof FactoryException) {
					throw (FactoryException) e.getTargetException();
				} else if (e.getTargetException() instanceof ENodeException) {
					throw (ENodeException) e.getTargetException();
				} else {
					final StringBuffer errorMessage = new StringBuffer("\n"); //$NON-NLS-1$
					if (e.getTargetException() != null) {
						errorMessage.append(e.getCause().getClass().getName() + " : " + e.getCause().getMessage()); //$NON-NLS-1$
					} else {
						errorMessage.append(""); //$NON-NLS-1$
					}
					throw new ENodeException(
							AcceleoEngineMessages.getString("EvalJavaService.RuntimeException", new Object[] { getDisplayString(m), errorMessage.toString(), }), new Int2(call.getBegin(), //$NON-NLS-1$
									call.getEnd()), call.getTemplate(), node, true, e.getTargetException());
				}
			}
		} finally {
			if (SELECT.equals(call.getName()) && call.getArguments().size() > 0) {
				evalContext.getCurrentScriptContext().contextPop(ScriptContext.ARGUMENT_POSITION);
			}
		}

		ENode createTry = ENode.createTry(result, node);
		if (createTry != null) {
			return createTry;
		} else {
			throw new ENodeException(AcceleoEngineMessages.getString("EvalJavaService.RuntimeType", new Object[] { getDisplayString(m), result.getClass().getName(), }), new Int2(call.getBegin(), call //$NON-NLS-1$
					.getEnd()), call.getTemplate(), node, true);
		}

	}

	/**
	 * Gives the method that can be invoked if any or null otherwise.
	 * 
	 * @param name
	 *            the name of the method
	 * @param argTypes
	 *            types of arguments
	 * @param receiver
	 *            the receiver of the result
	 * 
	 * @return the method with the given name and agruments type that return a
	 *         result assignable to the receiver
	 */
	private Method eGetMethod(String name, Class<?>[] argTypes, Object receiver) {
		Method[] methods = getPotentialMethods(name, argTypes.length);
		if (methods != null) {
			for (int i = 0; i < methods.length; i++) {
				Class<?>[] parameterTypes = methods[i].getParameterTypes();
				// ASSERT parameterTypes.length > 0
				if (argTypes[0] != EObject.class || parameterTypes[0].isInstance(receiver)) {
					boolean ok = true;
					for (int j = 0; j < parameterTypes.length && ok; j++) {
						ok = !((argTypes[j] == EObject.class && !argTypes[j].isAssignableFrom(parameterTypes[j])) || (argTypes[j] != EObject.class && argTypes[j] != parameterTypes[j]));
					}
					if (ok) {
						return methods[i];
					}
				}
			}
		}
		return null;
	}

	/**
	 * {@inheritDoc}
	 */
	public Object resolveType(Object type, Call call, int depth) {
		if (instance != null && type != null) { // Remark : type == GENERIC_TYPE
			// => type != null
			List<Object> quickKey = new ArrayList<Object>();
			quickKey.add(type);
			quickKey.add(call.toString()); // not getLink() because arguments
			if (quickResolveType.containsKey(quickKey)) {
				return quickResolveType.get(quickKey);
			} else {
				Object resolvedType = null;
				// 'filter' case
				if (instance instanceof ENodeServices || instance instanceof EObjectServices) {
					if (getTypeResolveList().contains(call.getName()) && call.getArguments().size() > 0) {
						Expression expression = call.getArguments().get(0);
						if (expression instanceof StringLiteral) {
							Object result = resolveAsEClassifier((StringLiteral) expression);
							if (result != null) {
								resolvedType = result;
							}
						}
					}
				}
				if (resolvedType == null && (instance instanceof RequestServices || instance instanceof ENodeServices || instance instanceof ContextServices)) {
					if (getTypeBridgeList().contains(call.getName())) {
						resolvedType = type;
					}
				}
				// Use declared methods
				if (resolvedType == null) {
					Method[] methods = getPotentialMethods(call.getName(), call.getArguments().size() + 1);
					if (methods != null && methods.length > 0 && methods[0].getDeclaringClass() == instance.getClass()) {
						resolvedType = new Object();
					}
				}
				quickResolveType.put(quickKey, resolvedType);
				return resolvedType;
			}
		} else {
			return null;
		}
	}

	/**
	 * Resolves this literal as an EClassifier link.
	 * 
	 * @return an EClassifier or null if not found
	 */
	public EClassifier resolveAsEClassifier(StringLiteral literal) {
		String nsURI = literal.getTemplate().getName();
		EPackage metamodel = EPackage.Registry.INSTANCE.getEPackage(nsURI);
		if (metamodel == null) {
			metamodel = ETools.ecore2EPackage(nsURI, true);
		}
		if (literal.getValue() != null && literal.getTemplate() != null && metamodel != null) {
			String id = literal.getValue();
			EClassifier result = ETools.getEClassifier(metamodel, id);
			if (result == null) {
				result = ETools.getEClassifier(EcorePackage.eINSTANCE, id);
			}
			return result;
		} else {
			return null;
		}
	}

	/**
	 * {@inheritDoc}
	 */
	public Method[] getCompletionProposals(Object type, int depth) {
		if (instance != null) {
			Collection<Method> result = new TreeSet<Method>(new Comparator<Method>() {
				public int compare(Method m0, Method m1) {
					return (m0.getName() + m0.toString()).compareTo(m1.getName() + m1.toString());
				}
			});
			Method[] methods = instance.getClass().getMethods();
			for (int i = 0; i < methods.length; i++) {
				Method method = methods[i];
				if (method.getDeclaringClass() == instance.getClass() && method.getParameterTypes().length >= 1) {
					result.add(method);
				}
			}
			return result.toArray(new Method[] {});
		} else {
			return new Method[] {};
		}
	}

	/**
	 * {@inheritDoc}
	 */
	public boolean validateCall(Call call) {
		return "".equals(call.getPrefix()) || TemplateConstants.LINK_PREFIX_JAVA.equals(call.getPrefix()); //$NON-NLS-1$
	}

}
