/*
 * Copyright (c) 2008 Obeo.
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Obeo - initial API and implementation
 */
package org.acceleo.engine.template.impl.spec;

import fr.obeo.acceleo.gen.template.eval.ENode;
import fr.obeo.acceleo.gen.template.eval.ENodeCastException;
import fr.obeo.acceleo.gen.template.eval.ENodeException;

import org.acceleo.template.core.Template;
import org.acceleo.template.statements.Feature;
import org.acceleo.template.statements.For;
import org.acceleo.template.statements.If;
import org.acceleo.template.statements.Statement;
import org.acceleo.template.statements.Text;

import java.util.Iterator;

import org.acceleo.engine.AcceleoEngineMessages;
import org.acceleo.engine.context.EvalContext;
import org.acceleo.engine.context.ScriptContext;
import org.acceleo.tools.factory.FactoryException;
import org.acceleo.tools.strings.Int2;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;

/**
 * A tool class for statements block manipulation.
 * 
 * @author <a href="mailto:yvan.lussaud@obeo.fr">Yvan Lussaud</a>
 * 
 */
public final class EvaluableTools {
	/** Hidden constructor. */
	private EvaluableTools() {
	}

	/**
	 * Generation of sub-statements.
	 * 
	 * @param evaluable
	 *            the element owning the statements
	 * @param root
	 *            Current ENode
	 * @param statements
	 *            List of statements
	 * @param evalContext
	 *            Current evaluation context
	 * 
	 * @return An ENode for the statements
	 * @throws ENodeException
	 *             if current statement could not be generated
	 * @throws FactoryException
	 *             if current statement could not be generated
	 */
	public static ENode genStatements(IRecursiveCheckable evaluable, ENode root, EList<? extends Statement> statements, EvalContext evalContext) throws ENodeException, FactoryException {
		EObject eObj;
		try {
			eObj = root.getEObject();
		} catch (ENodeCastException e) {
			eObj = ENode.EMPTY;
		}

		// TODO startDebug(evalContext.getLaunchManager());
		// TODO stepDebugInput(evalContext.getLaunchManager(), eObj);

		try {
			ENode node = new ENode(ENode.EMPTY, eObj, evaluable, evalContext.getLaunchManager().isSynchronize());
			boolean parentIsStatement = evaluable instanceof For || evaluable instanceof If;
			// boolean emptyEvaluation;
			if ((!parentIsStatement) && evaluable.isRecursive(eObj)) {
				// emptyEvaluation = true;
				throw new ENodeException(
						AcceleoEngineMessages.getString("Template.RecursiveCall", new Object[] { evaluable.toString(), }), new Int2(evaluable.getBegin(), evaluable.getEnd()), evaluable.getTemplate(), eObj, true); //$NON-NLS-1$
			} else {
				ENode current = new ENode(eObj, evaluable, evalContext.getLaunchManager().isSynchronize());
				evalContext.getCurrentScriptContext().contextPush(ScriptContext.CURRENT_NODE, current);
				evalContext.getCurrentScriptContext().contextPush(ScriptContext.TEMPLATE_NODE, current);
				try {
					if (!parentIsStatement) {
						evaluable.getVisitedEObjects().add(eObj);
					}
					Statement precedingElement = null;
					ENode precedingChild = null;
					Iterator<? extends Statement> elements = statements.iterator();
					while (elements.hasNext()) {
						Statement element = elements.next();
						try {
							boolean isLastText = !elements.hasNext() && element instanceof Text;
							boolean isEObjectNode = node.isEObject() || (node.isList() && node.getList().size() > 0 && node.getList().get(0).isEObject());
							if (isLastText && isEObjectNode && ((IEvaluable) element).gen(root, evalContext).asString().trim().length() == 0) {
								break;
							}
						} catch (ENodeCastException e) {
							// never catch
						}
						ENode child = ((IEvaluable) element).gen(root, evalContext);
						try {
							Template elemTemplate = element.getTemplate();
							boolean isPrecededByNonEmptyText = precedingElement instanceof Text && precedingChild.isString() && precedingChild.getString().length() > 0;
							if (elemTemplate != null /*
														 * &&
														 * elemTemplate.getFile() !=
														 * null &&
														 * element.getScript().getFile().getName().endsWith(".mt")
														 * //$NON-NLS-1$
														 */
									&& element instanceof Feature && isPrecededByNonEmptyText) {
								String text = child.asString();
								if (text.indexOf("\n") > -1) { //$NON-NLS-1$
									callInternalIndent(child, precedingChild);
								}
							}
						} catch (ENodeCastException e) {
							// never catch
						}
						node.append(child);
						precedingElement = element;
						precedingChild = child;
					}
				} finally {
					evalContext.getCurrentScriptContext().contextPop(ScriptContext.CURRENT_NODE);
					evalContext.getCurrentScriptContext().contextPop(ScriptContext.TEMPLATE_NODE);
					if (!parentIsStatement) {
						evaluable.getVisitedEObjects().remove(eObj);
					}
				}
			}

			// emptyEvaluation = (node.size() == 0);
			// stepDebugOutput(evalContext.getLaunchManager(), eObj, node);
			return node;
		} finally {
			// endDebug(evalContext.getLaunchManager());
		}
	}

	/**
	 * Call internal indentation if needed.
	 * 
	 * @param child
	 *            the current child of the generation
	 * @param precedingChild
	 *            the last child of the generation
	 * @throws ENodeCastException
	 *             if preceding child is not adaptable to String
	 */
	private static void callInternalIndent(ENode child, ENode precedingChild) throws ENodeCastException {
		String precedingText = precedingChild.getString();
		int iLine = precedingText.lastIndexOf("\n"); //$NON-NLS-1$
		if (iLine > -1 && ((iLine + 1) < precedingText.length())) {
			String indent = precedingText.substring(iLine + 1);
			if (indent.trim().length() == 0) {
				child.stringCall("internalIndent:" + indent, 0, 0); //$NON-NLS-1$
			}
		}
	}

	/**
	 * {@inheritDoc}
	 */
	public static String toString(EList<? extends Statement> statements) {
		StringBuffer text = new StringBuffer(""); //$NON-NLS-1$
		if (statements != null) {
			for (Statement statement : statements) {
				text.append(statement.toString());
			}
		}
		return text.toString();
	}
}
