/*
 * Copyright (c) 2005-2008 Obeo.
 * 
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Obeo - initial API and implementation
 */
package org.acceleo.gen.extension;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.acceleo.gen.extension.chain.DefaultPostGeneration;
import org.acceleo.gen.extension.chain.IPostGeneration;
import org.acceleo.gen.extension.phantom.IPhantomExtension;
import org.acceleo.gen.extension.vm.DefaultTemplateLoader;
import org.acceleo.gen.extension.vm.IOverdoneCodeStrategy;
import org.acceleo.gen.extension.vm.ITemplateLoader;
import org.acceleo.tools.factory.ETools;
import org.eclipse.emf.ecore.resource.ResourceSet;

/**
 * Dummy class for an easy access to Acceleo extension capabilities.
 * 
 * @author <a href="mailto:yvan.lussaud@obeo.fr">Yvan Lussaud</a>
 * 
 */
public final class Extension {

	/**
	 * Storage for the template loader.
	 */
	private static ITemplateLoader templateLoader = new DefaultTemplateLoader();

	/**
	 * Overdone code strategy storage.
	 */
	private static Map<String, IOverdoneCodeStrategy> overdoneStrategies;

	/**
	 * Storage for phantom extension.
	 */
	private static List<IPhantomExtension> phantomExtension;

	/**
	 * Storage for post generation actions.
	 */
	private static List<IPostGeneration> postGenerationActions;

	/**
	 * ClassLoader to use.
	 */
	private static ClassLoader classLoader;

	/**
	 * Private constructor.
	 */
	private Extension() {
		// nothing to do here
	}

	/**
	 * Add an overdone strategy for the given file extension.
	 * 
	 * @param fileExt
	 *            the file extension
	 *            <p>
	 *            "*" is the default match
	 * @param strategy
	 *            the strategy
	 */
	public static void addOverdoneStrategy(String fileExt, IOverdoneCodeStrategy strategy) {
		if (overdoneStrategies == null) {
			overdoneStrategies = new HashMap<String, IOverdoneCodeStrategy>();
		}
		overdoneStrategies.put(fileExt, strategy);
	}

	/**
	 * Return the overdone strategy for the given file extension.
	 * 
	 * @param fileExt
	 *            the file extension
	 * @return the strategy matched by the extension or null if none are found
	 */
	public static IOverdoneCodeStrategy getOverdoneStrategy(String fileExt) {
		IOverdoneCodeStrategy ret = null;
		if (overdoneStrategies != null) {
			ret = overdoneStrategies.get(fileExt);
		}
		return ret;
	}

	/**
	 * Add a post generation action.
	 * 
	 * @param postGeneration
	 *            the post generation action to add
	 * @return true if added
	 */
	public static boolean addPostGenerationAction(IPostGeneration postGeneration) {
		if (postGenerationActions == null) {
			initPostGenerationActions();
		}
		return postGenerationActions.add(postGeneration);
	}

	/**
	 * Getter for the post generation action list.
	 * 
	 * @return the post generation action list
	 */
	public static List<IPostGeneration> getPostGenerationActions() {
		if (postGenerationActions == null) {
			initPostGenerationActions();
		}
		return postGenerationActions;
	}

	/**
	 * Initialization of the post generation action list with the default post
	 * generation action.
	 * 
	 * @return true if the initialization has been done properly
	 */
	private static boolean initPostGenerationActions() {
		postGenerationActions = new LinkedList<IPostGeneration>();
		return postGenerationActions.add(new DefaultPostGeneration());
	}

	/**
	 * Getter for the template loader.
	 * 
	 * @return the template loader
	 */
	public static ITemplateLoader getTemplateLoader() {
		return templateLoader;
	}

	/**
	 * Setter for the template loader.
	 * 
	 * @param templateLoader
	 *            the new template loader
	 */
	public static void setTemplateLoader(ITemplateLoader templateLoader) {
		Extension.templateLoader = templateLoader;
	}

	/**
	 * Add a IPhantomExtension.
	 * 
	 * @param phantomExt
	 *            IPhantomExtension to add
	 * @return true (as specified by Collection.add(E))
	 */
	// need CChain to be done
	// TODO YLU extension - IChainCustomAction+ - customId
	// TODO YLU extension - IChainPreCondition+ -
	public static boolean addPhantomExtension(IPhantomExtension phantomExt) {
		if (phantomExtension == null) {
			initPhantomExtension();
		}
		return phantomExtension.add(phantomExt);
	}

	/**
	 * Phantom extensions list.
	 * 
	 * @return phantom extensions list
	 */
	public static List<IPhantomExtension> getPhantomExtension() {
		if (phantomExtension == null) {
			initPhantomExtension();
		}
		return phantomExtension;
	}

	/**
	 * Initialization of the phantom extension list.
	 * 
	 * @return true if the initialization has been done properly
	 */
	private static boolean initPhantomExtension() {
		phantomExtension = new LinkedList<IPhantomExtension>();
		return true;
		// TODO check the default phantom
		// return phantomExtension.add(...);
	}

	/**
	 * Getter for the resource set.
	 * 
	 * @return the resourceSet
	 */
	public static ResourceSet getResourceSet() {
		return ETools.getResourceSet();
	}

	public static void setResourceSet(ResourceSet resourceSet) {
		ETools.setResourceSet(resourceSet);
	}

	public static ClassLoader getClassLoader() {
		if (classLoader == null) {
			classLoader = Thread.currentThread().getContextClassLoader();
		}
		return classLoader;
	}

	public static void setClassLoader(ClassLoader classLoader) {
		Extension.classLoader = classLoader;
	}

}
